"""Conformers geometry-related functions, primarily an RMSD sieve implementation.

This module provides an implementation of RMSD sieve, allowing for easy mathematical
comparision of conformers' geometry and filtering out similar ones, based on
user-provided "threshold of similarity".
"""
import math
from typing import Iterable, Iterator, Sequence, Union

import numpy as np


def find_atoms(
    atoms: Union[Sequence[int], np.ndarray],
    find: Union[int, Iterable[int], np.ndarray],
    reverse: bool = False,
) -> np.ndarray:
    """Get indices of wanted atoms.

    Parameters
    ----------
    atoms : Sequence of int or numpy.ndarray
        List of atoms represented by their atomic numbers.
    find : int, Sequence of int, or numpy.ndarray
        Element or list of elements, represented by their atomic numbers, which indices
        should be find in *atoms* array.
    reverse : bool
        If ``True``, indices of atoms NOT specified in *find* will be returned.

    Returns
    -------
    numpy.ndarray
        Indices of found elements.
    """
    if isinstance(find, Iterable):
        blade = np.isin(atoms, find)
    else:
        blade = np.equal(atoms, find)
    blade = np.logical_xor(blade, reverse)
    indices = np.where(blade)[0]  # *atoms* is assumed to be one-dimensional
    return indices


def select_atoms(
    values: Union[Sequence, np.ndarray],
    indices: Union[Sequence[int], np.ndarray],
) -> np.ndarray:
    """Filter given values to contain values only corresponding to atoms on given
    indices. Recognizes if given values are a list of values for one or for many
    conformers, but it must be in shape (A, N) or (C, A, N) respectively.
    """
    # TODO: add support for indexing with indices shaped (1, A) and (C, A).
    if not np.asanyarray(indices).size:
        output = np.array([])
    else:
        try:
            output = np.take(values, indices, 1)
        except np.AxisError:
            output = np.take(values, indices, 0)
    return output


def take_atoms(
    values: Union[Sequence, np.ndarray],
    atoms: Union[Sequence[int], np.ndarray],
    wanted: Union[int, Iterable[int], np.ndarray],
) -> np.ndarray:
    """Filters given values, returning those corresponding to atoms specified
    as *wanted*. Roughly equivalent to:
    >>> numpy.take(values, numpy.nonzero(numpy.equal(atoms, wanted))[0], 1)
    but returns empty array, if no atom in *atoms* matches *wanted* atom.
    If *wanted* is list of elements, numpy.isin is used instead of numpy.equal.

    Parameters
    ----------
    values: Sequence or numpy.ndarray
        array of values; it should be one-dimensional list of values or
        n-dimensional array of shape
        (conformers, values[, coordinates[, other]])
    atoms: Sequence of int or numpy.ndarray
        list of atoms in molecule, given as atomic numbers;
        order should be the same as corresponding values for each conformer
    wanted: int or Iterable of int or numpy.ndarray
        atomic number of wanted atom, or a list of those

    Returns
    -------
    numpy.ndarray
        values trimmed to corresponding to desired atoms only; preserves
        original dimension information
    """
    indices = find_atoms(atoms, wanted)
    return select_atoms(values, indices)


def drop_atoms(
    values: Union[Sequence, np.ndarray],
    atoms: Union[Iterable[int], np.ndarray],
    discarded: Union[int, Iterable[int], np.ndarray],
) -> np.ndarray:
    """Filters given values, returning those corresponding to atoms not
    specified as discarded. Roughly equivalent to:
    >>> numpy.take(values, numpy.nonzero(~numpy.equal(atoms, discarded))[0], 1)
    If *wanted* is list of elements, numpy.isin is used instead of numpy.equal.

    Parameters
    ----------
    values: Sequence or numpy.ndarray
        array of values; it should be one-dimensional list of values or
        n-dimensional array of shape
        (conformers, values[, coordinates[, other]])
    atoms: Iterable of int or numpy.ndarray
        list of atoms in molecule, given as atomic numbers;
        order should be the same as corresponding values for each conformer
    discarded: int or Iterable of int or numpy.ndarray
        atomic number of discarded atom, or a list of those

    Returns
    -------
    numpy.ndarray
        values trimmed to corresponding to desired atoms only; preserves
        original dimension information
    """
    indices = find_atoms(atoms, discarded, reverse=True)
    return select_atoms(values, indices)


def is_triangular(n: int) -> bool:
    """Checks if number *n* is triangular.

    Notes
    -----
    If *n* is the mth triangular number, then n = m*(m+1)/2.
    Solving for m using the quadratic formula: m = (sqrt(8n+1) - 1) / 2,
    so *n* is triangular if and only if 8n+1 is a perfect square.

    Parameters
    ----------
    n: int
        number to check

    Returns
    -------
    bool
        True if number *n* is triangular, else False
    """
    if n < 0:
        return False
    check = math.sqrt(8 * n + 1)
    try:
        return check == int(check)
    except OverflowError:
        # check is float('inf')
        return False


def get_triangular_base(n: int) -> int:
    """Find which mth triangular number *n* is."""
    if not is_triangular(n):
        raise ValueError(f'"n" should be a triangular number. {n} is not triangular.')
    return int((math.sqrt(8 * n + 1) - 1) / 2)


def get_triangular(m: int) -> int:
    """Find *m*\\th triangular number."""
    if m < 0:
        raise ValueError('"m" should be non-negative number.')
    if not m // 1 == m:
        raise ValueError(f'"m" should be a whole number, {m} given.')
    return m * (m + 1) // 2


MoleculeOrList = Union[Sequence[Sequence[float]], Sequence[Sequence[Sequence[float]]]]


def center(a: MoleculeOrList) -> MoleculeOrList:
    """Zero-center all given conformers by subtracting their centroids.
    Accepts single molecule or list of conformers."""
    a = np.asanyarray(a)
    return a - np.expand_dims(a.mean(axis=-2), -2)


def kabsch_rotate(a: MoleculeOrList, b: MoleculeOrList) -> np.ndarray:
    """Minimize RMSD of conformers *a* and *b* by rotating molecule *a* onto *b*.
    Expects given representation of conformers to be zero-centered.
    Both *a* and *b* may be a single molecule or a set of conformers.

    Parameters
    ----------
    a : [Sequence of ]Sequence of Sequence of float
        Set of points representing atoms, that will be rotated to best match reference.
    b : [Sequence of ]Sequence of Sequence of float
        Set of points representing atoms of the reference molecule.

    Returns
    -------
    numpy.ndarray
        Rotated set of points *a*.

    Notes
    -----
    Uses Kabsch algorithm, also known as Wahba's problem. See:
    https://en.wikipedia.org/wiki/Kabsch_algorithm and
    https://en.wikipedia.org/wiki/Wahba%27s_problem
    """
    # this implementation uses Einstein summation convention, numpy.eisnum()
    # this approach is probably not the most efficient
    # but lets us easily perform a matrix multiplication on stacks of matrices
    a, b = np.asanyarray(a), np.asanyarray(b)
    # calculate covariance matrix for each stacked set of points
    # for each of stacked sets of points, equivalent of:
    # >>> cov = a.T @ b
    cov = np.einsum("...ji,...jk", a, b)
    u, s, vh = np.linalg.svd(cov)  # singular value decomposition
    # if determinant is negative, swap to ensure right-handed coordinate system
    det = np.linalg.det(vh @ u)  # works with stacked matrices
    # don't introduce new dimension if not necessary
    shape = (det.size, 3, 3) if det.size > 1 else (3, 3)
    swap = np.zeros(shape)
    swap[..., np.arange(2), np.arange(2)] = 1
    swap[..., -1, -1] = np.sign(det)
    # calculate optimally rotated set/s of points `a`
    # for each of stacked sets of points, equivalent of
    # >>> rotated = a @ u @ swap @ vh
    # where u @ swap @ vh is rotation matrix
    return np.einsum("...ij,...jk,...kl,...lm", a, u, swap, vh)


def calc_rmsd(a: MoleculeOrList, b: MoleculeOrList) -> np.ndarray:
    """Compute RMSD (round-mean-square deviation) of two conformers (or sets of them).

    Parameters
    ----------
    a : [Sequence of ]Sequence of Sequence of float
        Set of points representing atoms or list thereof.
    b : [Sequence of ]Sequence of Sequence of float
        Set of points representing atoms or list thereof.

    Returns
    -------
    float or numpy.ndarray
        Value of RMSD of two conformers or list of values, if list of conformers given.

    Notes
    -----
    https://en.wikipedia.org/wiki/Root-mean-square_deviation_of_atomic_positions
    """
    deviation = np.asanyarray(a) - np.asanyarray(b)
    # get a sum of two last dimensions by using `axis=(-2, -1)`
    return np.sqrt(np.square(deviation).sum(axis=(-2, -1)) / deviation.shape[-2])


def fixed_windows(series: Sequence, size: int) -> np.ndarray:
    """Simple, vectorized implementation of basic sliding window.
    Produces a list of windows of given *size* from given *series*.

    Parameters
    ----------
    series : sequence
        Series of data, of which sliding window view is requested.
    size : int
        Number of data points in the window. Must be a positive integer.

    Returns
    -------
    numpy.ndarray
        List of indices, corresponding to values in the original array,
        that form a window

    Raises
    ------
    ValueError
        if non-positive integer given as window size
    TypeError
        if non-integer value given as window size

    Notes
    -----
    Implementation inspired by
    https://towardsdatascience.com/fast-and-robust-sliding-window-vectorization-with-numpy-3ad950ed62f5
    """
    if not isinstance(size, int):
        raise TypeError(f"*size* must be a positive integer, but {type(size)} given.")
    elif size <= 0:
        raise ValueError(f"*size* must be a positive integer, but {size} given.")
    series = np.asanyarray(series)
    # create indices for fancy indexing [[0, 1, ..., size], [1, 2, ..., size + 1], ...]
    windows = (
        np.arange(size)[np.newaxis, ...]
        + np.arange(series.size - size + 1)[np.newaxis, ...].T
    )
    return windows


def stretching_windows(
    values: Sequence[float],
    size: Union[int, float],
    keep_hermits: bool = False,
    hard_bound: bool = False,
) -> Iterator[np.ndarray]:
    """Implements a sliding window of a variable size, where values in each window are
    at most *size* bigger than the lowest value in given window. Values yielded
    are :class:`np.ndarray`\\s of indices of sorted values, that constitute each window.

    When window reaches a border, that is an end of the *values* array or a gap between
    values that is larger than given *size*, it is "squeezed", when pressed against the
    border, producing subsequences of the first view that touches a border. This is
    usefull, when one wants to form a window for each value in the original array.

    >>> list(stretching_windows([1, 2, 3, 4, 7, 8], 3))
    [[0, 1, 2], [1, 2, 3], [2, 3], [4, 5]]

    This "soft" right bound may be "hardened" by passing ``hard_bound=True`` as a
    parameter to a function call. A window will than move immediately to the border's
    other side.

    >>> list(stretching_windows([1, 2, 3, 4, 7, 8], 3), hard_bound=True)
    [[0, 1, 2], [1, 2, 3], [4, 5]]

    Windows of size 1, called hermits, are by default ignored.

    >>> arr = [1, 2, 10, 20, 22]
    >>> list(stretching_windows(arr, 5))
    [[0, 1], [3, 4]]

    If such behavior is not desired, it may be turned off with ``keep_hermits = True``.
    One must remember that, when a bound is "soft", the last window is always a hermit.

    >>> list(stretching_windows(arr, 5, keep_hermits=True))
    [[0, 1], [1], [2], [3, 4], [4]]
    >>> list(stretching_windows(arr, 5, keep_hermits=True, hard_bound=True))
    [[0, 1], [2], [3, 4]]

    Parameters
    ----------
    values : Sequence of float
        List of values, on which sliding window view is requested.
    size : int or float
        Maximum difference of smallest and largest values inside each window.
    keep_hermits : bool
        If windows of size one should be yielded (True) or omitted (False).
        False by default.
    hard_bound : bool
        How window should behave close to borders. With hard bound (True) it will move
        to the other side of border as soon, as it is reached. With soft bound (False)
        it will "squeeze" when pressed against the border, producing subsequences of
        the first view that includes border value. False by default.

    Yields
    ------
    np.array of int
        List of indices, corresponding to sorted values in the original array,
        that form a window.

    Raises
    ------
    ValueError
        If given *size* is not a positive number.
    """
    if size <= 0:
        raise ValueError("Size of the energy window must be a positive number.")
    order = np.argsort(values)
    if not order.size:
        return
    ordered = np.asanyarray(values)[order]
    # side="right" is ie. "or equal to" part of "include lower on equal to value+size"
    indices = np.searchsorted(ordered, ordered + size, side="right")
    bounds = np.stack([np.arange(indices.size), indices], axis=1)  # shape (N, 2)
    if hard_bound:
        # don't allow repeated stop index
        unique = np.insert(np.diff(indices).astype(bool), 0, True)
        bounds = bounds[unique]
    if not keep_hermits:
        # if difference between start and stop is 1, it is a hermit
        hermits = np.diff(bounds, axis=1) == 1
        bounds = bounds[~hermits.reshape(-1)]
    for start, stop in bounds:
        # if necessary make it order[start:] to include last value
        yield order[start : (stop if stop <= indices.size else None)]


def pyramid_windows(series: Sequence) -> Iterator[np.ndarray]:
    """Produces windows of shrinking sizes, from full sequence to last element only.

    This function yields :class:`numpy.ndarray`\\s with indices that may be used to
    index an original sequence (assuming original sequence is :class:`numpy.ndarray` as
    well). The first window yielded represents a whole *series* sequence and each
    consecutive window is reduced by the first element, leaving only the last element in
    the final window. This allows for easy setup of efficient calculations in symmetric
    each-to-each relationship.

    >>> series = [3, 6, 3, 5, 7]
    >>> for window in pyramid_windows(series):
    ...     print(window)
    [0 1 2 3 4]
    [1 2 3 4]
    [2 3 4]
    [3 4]
    [4]

    Parameters
    ----------
    series : sequence
        Sequence of elements, for which windows should be generated.

    Yields
    ------
    np.ndarray(dtype=int)
        Windows as np.ndarray of indices.
    """
    length = len(series)
    for idx in range(length):
        yield np.arange(start=idx, stop=length, step=1, dtype=int)


def rmsd_sieve(
    geometry: Sequence[Sequence[Sequence[float]]],
    windows: Iterable[Sequence[int]],
    threshold: float = 1,
) -> np.ndarray:
    """Compare conformers' geometry to keep only those that differ at least by a given
    threshold.

    This function calculates how similar conformers are one to another, using a RMSD
    measure, that is is a root-mean-square deviation of atomic positions, and signalizes
    which of the conformers are duplicates, according to a given similarity threshold.
    Returned array of booleans may be treated as "originality" indicators for each
    conformer: ``True`` means given conformer has distinct structure, ``False`` means
    given conformer is similar to some other conformer marked as "original".

    The measure of conformers' similarity, the *threshold* parameter, is a minimum value
    of RMSD needed to consider two conformers different. In other words, if two
    conformers give a RMSD value that is lower then *threshold*, one of them will be
    marked as similar, producing a ``False`` in the output array.

    To lower a computational expense, similarity measurement is performed in "chunks",
    using a sliding window technique. Windows consist of a portion of conformers from
    the original data, or more precisely, indices of conformers that should be included
    in the particular window. First item from the window is compared to all the others
    that are in the same window, and if any of them is similar to the reference item,
    it is marked as duplicate (not "original"). The process is repeated for each window.

    The windows itself should be provided by user as *windows* parameter. This provides
    a flexibility in the process: you may choose to sacrifice accuracy to lower
    necessary computational time or vice versa. You may also choose a different moving
    window strategy or reject it alltogether, and calculate one-to-each similarity in
    the whole set. Iterables of windows accepted by this function may be generated with
    one of the dedicated moving window funcions: :func:`.stretching_windows`,
    :func:`.fixed_windows`, or :func:`.pyramid_windows`. Refer to their documentation
    for more information.

    Parameters
    ----------
    geometry : sequence of sequence of sequence of float
        A list of conformers, where each conformer is represented by a sequence of
        coordinates in 3-dimensional space. It is assumed that order of atoms in each
        conformers' representation is identical.
    windows : iterable of sequence of int
        An iterable of windows, where each window is a list of indices. Comparision of
        RMSD values will be performed inside each window.
    threshold : float
        Minimum RMSD value to consider two compared conformers different.

    Returns
    -------
    np.ndarray(dtype=bool)
        Array of booleans for each conformer: ``True`` if conformer's structure is
        "original" and should be kept, ``False`` if it is a duplicate of other,
        "original" structure (at least according to *threshold* given), and should be
        discarded.
    """
    blade = np.ones(len(geometry), dtype=bool)
    # zero-center all conformers
    geometry = center(geometry)
    for window in windows:
        # don't include values already discarded
        reduced_window = window[blade[window]]
        if reduced_window.size <= 1:
            # only one molecule in the window, we keep it
            # or no conformers at all
            continue
        head, *tail = geometry[reduced_window]  # reference, other
        # find best rotation of mols in window onto first mol
        tail = kabsch_rotate(tail, head)
        # calculate RMSD list of mols to first mol
        rmsd = calc_rmsd(head, tail)
        # if RMSD <= threshold mark in blade as False, first one is always kept
        blade[reduced_window[1:]] = rmsd > threshold
    # return blade with True for each molecule kept
    return blade



"""Data containers."""

from .array_base import ArrayBase, ArrayProperty
from .arrays import (
    Bands,
    BooleanArray,
    DataArray,
    ElectronicActivities,
    ElectronicData,
    Energies,
    FilenamesArray,
    FloatArray,
    Geometry,
    InfoArray,
    IntegerArray,
    ScatteringActivities,
    ScatteringData,
    SpectralActivities,
    SpectralData,
    Transitions,
    VibrationalActivities,
    VibrationalData,
)
from .conformers import Conformers
from .spectra import SingleSpectrum, Spectra



"""Data export to CSV format."""
import csv
import logging as lgg
from contextlib import contextmanager
from itertools import repeat, zip_longest
from pathlib import Path
from string import Template
from typing import IO, Any, AnyStr, Dict, Iterable, Iterator, List, Optional, Union

import numpy as np

from ..glassware.arrays import (
    Bands,
    Energies,
    FloatArray,
    SpectralActivities,
    SpectralData,
    Transitions,
)
from ..glassware.spectra import SingleSpectrum, Spectra
from .writer_base import WriterBase, _GenericArray

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


class _CsvMixin:
    """Mixin class for csv writers.

    This class takes care of setting up format of produced csv files.
    It should be used as a first base class to ensure proper cooperation with other
    base classes. It will pass all given *args and **kwargs to the next base class
    in MRO.
    """

    _known_fmt_params = {
        "delimiter",
        "doublequote",
        "escapechar",
        "lineterminator",
        "quotechar",
        "quoting",
        "skipinitialspace",
        "strict",
    }

    def __init__(
        self,
        *args,
        dialect: Union[str, csv.Dialect] = "excel",
        fmtparams: Optional[Dict] = None,
        include_header: bool = True,
        **kwargs,
    ):
        """
        Parameters
        ----------
        dialect: str or csv.Dialect
            Name of a dialect or :class:`csv.Dialect` object, which will be used by
            :class:`.csv.writer`.
        fmtparams: dict, optional
            Additional formatting parameters for :class:`.csv.writer` to use.
            For list of valid parameters consult :class:`csv.Dialect` documentation.
        include_header: bool, optional
            Determines if file should contain a header with column names, ``True`` by
            default.
        """
        self.dialect = dialect
        self.fmtparams = fmtparams or {}
        self.include_header = include_header
        super().__init__(*args, **kwargs)

    @property
    def dialect(self):
        """Name of a dialect (as string) or csv.Dialect object,
        which will be used by csv.writer.
        """
        return self._dialect

    @dialect.setter
    def dialect(self, dialect: Union[str, csv.Dialect]):
        self._dialect = (
            dialect if isinstance(dialect, csv.Dialect) else csv.get_dialect(dialect)
        )

    @property
    def fmtparams(self):
        """Dict of additional formatting parameters for csv.writer to use.
        For list of valid parameters consult csv.Dialect documentation.

        Raises
        ------
        TypeError
            if invalid parameter is given
        """
        return self._fmtparams

    @fmtparams.setter
    def fmtparams(self, params: Dict):
        for param in params.keys():
            if param not in self._known_fmt_params:
                raise TypeError(f"'{param}' is an invalid csv formatting parameter")
        self._fmtparams = params


# CLASSES
class CsvWriter(_CsvMixin, WriterBase):
    """Writes extracted or calculated data to .csv format files."""

    extension = "csv"

    def __init__(
        self,
        destination: Union[str, Path],
        mode: str = "x",
        include_header: bool = True,
        dialect: Union[str, csv.Dialect] = "excel",
        **fmtparams,
    ):
        """
        Parameters
        ----------
        destination: str or pathlib.Path
            Directory, to which generated files should be written.
        mode: str
            Specifies how writing to file should be handled. Should be one of
            characters: 'a' (append to existing file), 'x' (only write if file doesn't
            exist yet), or 'w' (overwrite file if it already exists).
        include_header: bool, optional
            Determines if file should contain a header with column names, ``True`` by
            default.
        dialect: str or csv.Dialect
            Name of a dialect or :class:`csv.Dialect` object, which will be used by
            underlying :class:`csv.writer`.
        fmtparams: dict, optional
            Additional formatting parameters for underlying csv.writer to use.
            For list of valid parameters consult :class:`csv.Dialect` documentation.
        """
        super().__init__(
            destination=destination,
            mode=mode,
            dialect=dialect,
            fmtparams=fmtparams,
            include_header=include_header,
        )

    @contextmanager
    def _get_handle(
        self,
        template: Union[str, Template],
        template_params: dict,
        open_params: Optional[dict] = None,
    ) -> Iterator[IO[AnyStr]]:
        open_params = open_params or {"newline": ""}
        with super()._get_handle(template, template_params, open_params) as handle:
            yield handle

    def _iter_handles(
        self,
        filenames: Iterable[str],
        template: Union[str, Template],
        template_params: dict,
        open_params: Optional[dict] = None,
    ) -> Iterator[IO[AnyStr]]:
        open_params = open_params or {"newline": ""}
        yield from super()._iter_handles(
            filenames, template, template_params, open_params
        )

    def generic(
        self,
        data: List[_GenericArray],
        name_template: Union[str, Template] = "${cat}.${det}.${ext}",
    ):
        """Writes generic data from multiple :class:`.DataArray`-like objects to a
        single file. Said objects should provide a single value for each conformer.

        Parameters
        ----------
        data
            :class:`.DataArray` objects that are to be exported.
        name_template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        genres = [arr.genre for arr in data]
        header = ["Gaussian output file"] + [self._header[genre] for genre in genres]
        values = [arr.values for arr in data]
        rows = zip(data[0].filenames, *values)
        types = [type(arr).__name__.lower().replace("array", "") for arr in data]
        detail = "various" if len(set(types)) > 1 else types[0]
        genre = "misc" if len(genres) > 1 else genres[0]
        template_params = {
            "cat": "generic",
            "conf": "multiple",
            "det": detail,
            "genre": genre,
        }
        with self._get_handle(name_template, template_params) as handle:
            csvwriter = csv.writer(handle, dialect=self.dialect, **self.fmtparams)
            if self.include_header:
                csvwriter.writerow(header)
            for row in rows:
                csvwriter.writerow(row)

    def energies(
        self,
        energies: Energies,
        corrections: Optional[FloatArray] = None,
        name_template: Union[str, Template] = "distribution-${genre}.${ext}",
    ):
        """Writes Energies object to csv file. The output also contains derived values:
        populations, min_factors, deltas. Corrections are added only when explicitly
        given.

        Parameters
        ----------
        energies: glassware.Energies
            Energies objects that is to be serialized
        corrections: glassware.DataArray, optional
            DataArray objects containing energies corrections
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        header = ["Gaussian output file"]
        header += "population min_factor delta energy".split(" ")
        if corrections is not None:
            header += ["corrections"]
            corr = corrections.values
        else:
            corr = []
        rows = zip_longest(
            energies.filenames,
            energies.populations,
            energies.min_factors,
            energies.deltas,
            energies.values,
            corr,
        )
        template_params = {
            "conf": "multiple",
            "genre": energies.genre,
            "cat": "populations",
        }
        with self._get_handle(name_template, template_params) as handle:
            csvwriter = csv.writer(handle, dialect=self.dialect, **self.fmtparams)
            if self.include_header:
                csvwriter.writerow(header)
            for row in rows:
                csvwriter.writerow(v for v in row if v is not None)
        logger.info("Energies export to csv files done.")

    def _energies_handler(self, data: List[Energies], extras: Dict[str, Any]) -> None:
        # TODO: return to WriterBase's implementation when `.overview()` added
        for en in data:
            self.energies(
                en, corrections=extras.get("corrections", dict()).get(en.genre)
            )

    def single_spectrum(
        self,
        spectrum: SingleSpectrum,
        name_template: Union[str, Template] = "${cat}.${genre}-${det}.${ext}",
    ):
        """Writes SingleSpectrum object to csv file.

        Parameters
        ----------
        spectrum: glassware.SingleSpectrum
            spectrum, that is to be serialized
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        template_params = {
            "genre": spectrum.genre,
            "cat": "spectrum",
            "det": spectrum.averaged_by,
        }
        with self._get_handle(name_template, template_params) as handle:
            csvwriter = csv.writer(handle, dialect=self.dialect, **self.fmtparams)
            if self.include_header:
                csvwriter.writerow([spectrum.units["y"], spectrum.units["x"]])
            for row in zip(spectrum.x, spectrum.y):
                csvwriter.writerow(row)
        logger.info("Spectrum export to csv files done.")

    def spectral_activities(
        self,
        band: SpectralActivities,
        data: List[SpectralActivities],
        name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
    ):
        """Writes SpectralActivities objects to csv files (one file for each conformer).

        Parameters
        ----------
        band: glassware.SpectralActivities
            Object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data.
        data: list of glassware.SpectralActivities
            SpectralActivities objects that are to be serialized; all should contain
            information for the same set of conformers and correspond to given band.
            Assumes that all *data*'s elements have the same *spectra_type*, which is
            passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        self._spectral(
            band=band,
            data=data,
            name_template=name_template,
            category="activities",
        )

    def spectral_data(
        self,
        band: SpectralData,
        data: List[SpectralData],
        name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
    ):
        """Writes SpectralData objects to csv files (one file for each conformer).

        Parameters
        ----------
        band: glassware.SpectralData
            Object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data.
        data: list of glassware.SpectralData
            SpectralData objects that are to be serialized; all should contain
            information for the same set of conformers and correspond to given band.
            Assumes that all *data*'s elements have the same *spectra_type*, which is
            passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        self._spectral(
            band=band, data=data, name_template=name_template, category="data"
        )

    def _spectral(
        self,
        band: SpectralActivities,
        data: Union[List[SpectralData], List[SpectralActivities]],
        name_template: Union[str, Template],
        category: str,
    ):
        """Writes SpectralData objects to csv files (one file for each conformer).

        Parameters
        ----------
        band: glassware.SpectralData
            Object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data.
        data: list of glassware.SpectralData
            SpectralData objects that are to be serialized; all should contain
            information for the same set of conformers and correspond to given band.
            Assumes that all *data*'s elements have the same *spectra_type*, which is
            passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        category : str
            category of exported data genres

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        try:
            spectra_type = data[0].spectra_type
        except IndexError:
            raise ValueError("No data to export.")
        data = [band] + data
        headers = [self._header[bar.genre] for bar in data]
        values = zip(*[bar.values for bar in data])
        template_params = {"genre": band.genre, "cat": category, "det": spectra_type}
        for values_, handle in zip(
            values,
            self._iter_handles(band.filenames, name_template, template_params),
        ):
            csvwriter = csv.writer(handle, dialect=self.dialect, **self.fmtparams)
            if self.include_header:
                csvwriter.writerow(headers)
            for row in zip(*values_):
                csvwriter.writerow(row)
        logger.info(f"{category.title()} export to csv files done.")

    def spectra(
        self,
        spectra: Spectra,
        name_template: Union[str, Template] = "${conf}.${genre}.${ext}",
    ):
        """Writes Spectra object to .csv files (one file for each conformer).

        Parameters
        ----------
        spectra: glassware.Spectra
            Spectra object, that is to be serialized.
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        abscissa = spectra.x
        header = [spectra.units["y"], spectra.units["x"]]
        template_params = {"genre": spectra.genre, "cat": "spectra"}
        for values, handle in zip(
            spectra.y,
            self._iter_handles(spectra.filenames, name_template, template_params),
        ):
            csvwriter = csv.writer(handle, dialect=self.dialect, **self.fmtparams)
            if self.include_header:
                csvwriter.writerow(header)
            for row in zip(abscissa, values):
                csvwriter.writerow(row)
        logger.info("Spectra export to csv files done.")

    def transitions(
        self,
        transitions: Transitions,
        wavelengths: Bands,
        only_highest=True,
        name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
    ):
        """Writes electronic transitions data to CSV files (one for each conformer).

        Parameters
        ----------
        transitions : glassware.Transitions
            Electronic transitions data that should be serialized.
        wavelengths : glassware.ElectronicActivities
            Object containing information about wavelength at which transitions occur.
        only_highest : bool
            Specifies if only transition of highest contribution to given band should
            be reported. If ``False`` all transition are saved to file.
            Defaults to ``True``.
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        transtions_data = (
            transitions.highest_contribution
            if only_highest
            else (
                transitions.ground,
                transitions.excited,
                transitions.values,
                transitions.contribution,
            )
        )
        header = ["wavelength/nm", "ground", "excited", "coefficient", "contribution"]
        template_params = {
            "genre": transitions.genre,
            "cat": "transitions",
            "det": "highest" if only_highest else "all",
        }
        for grounds, exciteds, values, contribs, bands, handle in zip(
            *transtions_data,
            wavelengths.wavelen,
            self._iter_handles(transitions.filenames, name_template, template_params),
        ):
            csvwriter = csv.writer(handle, dialect=self.dialect, **self.fmtparams)
            if self.include_header:
                csvwriter.writerow(header)
            for g, e, v, c, b in zip(grounds, exciteds, values, contribs, bands):
                try:
                    listed = [
                        d
                        for d in zip(repeat(b), g, e, v, c)
                        # omit entry if any value is masked
                        if all(x is not np.ma.masked for x in d)
                    ]
                except TypeError:
                    # transition_data is transitions.highest_contribution
                    listed = [(b, g, e, v, c)]
                for data in listed:
                    csvwriter.writerow(data)



import re

import pytest

from tesliper.extraction import gaussian_parser as gp


@pytest.mark.parametrize(
    "value",
    [
        "1",
        " 1",
        " -1",
        "23",
        "0.2",
        "0.243",
        "123.657",
        "-0.42",
        "-3425.42",
        ".92",
        "-.42",
    ],
)
def test_number_matches(value):
    assert re.search(gp.number, value)


@pytest.mark.xfail
@pytest.mark.parametrize("value", ["-", ".", "- 1", "12-", "42."])
def test_number_not_matches(value):
    assert not re.search(gp.number, value)


@pytest.mark.parametrize(
    "value",
    [
        "3e24",
        "3e-656",
        "3E24",
        "3E-24",
        "-3e24",
        "-3e-656",
        "-3E24",
        "-3E-24",
        "3.23e24",
        "3.23e-656",
        "3.23E24",
        "3.23E-24",
        "-3.23e24",
        "-3.23e-656",
        "-3.23E24",
        "-3.23E-24",
    ],
)
def test_number_sci_matches(value):
    assert re.search(gp.number, value)


@pytest.mark.xfail
@pytest.mark.parametrize(
    "value", ["42e", "42e-", "42.e", "42.e-", "42E", "42E-", "42.E", "42.E-"]
)
def test_number_sci_not_matches(value):
    assert not re.search(gp.number, value)


@pytest.mark.parametrize(
    "line",
    [
        " Stoichiometry    CH4O\n",
        " Stoichiometry    CH4O(1+)\n",
        " Stoichiometry    CH4O(1-)\n",
        " Stoichiometry    CH4O(1+,2)\n",
        " Stoichiometry    CH4O(1-,2)\n",
        " Stoichiometry    CH4O(3)\n",
    ],
)
def test_stoichiometry(line):
    assert gp.STOICH_CRE.match(line)



# IMPORTS
import functools
import logging as lgg
import queue
import tkinter as tk
import tkinter.ttk as ttk
from collections import Counter, namedtuple
from pathlib import Path
from tkinter import messagebox
from tkinter.filedialog import (
    askdirectory,
    askopenfilename,
    askopenfilenames,
    asksaveasfilename,
)

import numpy as np

from tesliper import Soxhlet
from tesliper import datawork as dw
from tesliper import glassware as gw
from tesliper import writing as wr

# LOGGER
from ...datawork import DEFAULT_ACTIVITIES
from .choices import ColorsChoice, ConformersChoice, EnergiesChoice
from .collapsible_pane import CollapsiblePane
from .helpers import ThreadedMethod, join_with_and
from .label_separator import LabelSeparator
from .numeric_entry import NumericEntry
from .popups import ExportPopup, GjfPopup

logger = lgg.getLogger(__name__)


# CLASSES
class FilterRange(ttk.Frame):
    def __init__(self, parent, view, proxy, **kwargs):
        super().__init__(parent, **kwargs)
        self.view = view
        # dict with getters for "genre" and "show" comboboxes
        # and "units" of currently showing values
        self.proxy = proxy

        self.columnconfigure(1, weight=1)
        self.lower_var = tk.StringVar()
        self.upper_var = tk.StringVar()
        self.units_var = tk.StringVar()
        ttk.Label(self, text="Minimum").grid(column=0, row=0)
        ttk.Label(self, text="Maximum").grid(column=0, row=1)
        ttk.Label(self, textvariable=self.units_var, width=8).grid(column=2, row=0)
        ttk.Label(self, textvariable=self.units_var, width=8).grid(column=2, row=1)
        ttk.Label(self, text="Energy type").grid(column=0, row=2)
        self.lower_entry = NumericEntry(
            self,
            textvariable=self.lower_var,
            width=15,
            scroll_modifier=self.scroll_modifier,
            keep_trailing_zeros=True,
            decimal_digits=6,
            rounding="down",
        )
        self.lower_entry.grid(column=1, row=0, sticky="new")
        self.upper_entry = NumericEntry(
            self,
            textvariable=self.upper_var,
            width=15,
            scroll_modifier=self.scroll_modifier,
            keep_trailing_zeros=True,
            decimal_digits=6,
            rounding="up",
        )
        self.upper_entry.grid(column=1, row=1, sticky="new")

        b_filter = ttk.Button(self, text="Trim to...", command=self.filter_energy)
        b_filter.grid(column=0, row=2, columnspan=3, sticky="new")

        # trigger action on Enter key press
        self.lower_entry.bind("<Return>", self.filter_energy, "+")
        self.upper_entry.bind("<Return>", self.filter_energy, "+")

        root = self.winfo_toplevel()
        # root.bind("<<KeptChanged>>", self.set_upper_and_lower, "+")
        root.bind("<<DataExtracted>>", self.set_upper_and_lower, "+")

        root.changer.register(
            [b_filter, self.lower_entry, self.upper_entry], "energies"
        )

    _scroll_modifiers = {
        "values": lambda v, d: v + 0.00001 * d,
        "deltas": lambda v, d: v + 0.01 * d,
        "min_factors": lambda v, d: v + 1.2 ** d,
        "populations": lambda v, d: v + 1 * d,
    }

    _entry_configure = {
        "values": {
            "min_value": float("-inf"),
            "max_value": float("inf"),
            "decimal_digits": 6,
        },
        "deltas": {"min_value": 0, "max_value": float("inf"), "decimal_digits": 4},
        "min_factors": {"min_value": 0, "max_value": float("inf"), "decimal_digits": 4},
        "populations": {"min_value": 0, "max_value": 100, "decimal_digits": 4},
    }

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    def on_show_selected(self, _event=None):
        config = self._entry_configure[self.proxy["show"]()]
        self.upper_entry.configure(**config)
        self.lower_entry.configure(**config)
        self.set_upper_and_lower()

    def scroll_modifier(self, value, delta):
        showing = self.proxy["show"]()
        updated = self._scroll_modifiers[showing](value, delta)
        return updated

    def set_upper_and_lower(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.set_upper_and_lower handler.")
        showing = self.proxy["show"]()
        units = self.proxy["units"]()
        factor = 100 if showing == "populations" else 1
        try:
            energy = self.proxy["genre"]()
            arr = getattr(self.tesliper[energy], showing)
            lower, upper = arr.min(), arr.max()
        except (KeyError, ValueError):
            lower, upper = 0, 0
        else:
            lower, upper = lower * factor, upper * factor
        finally:
            self.lower_entry.update(lower)
            self.upper_entry.update(upper)
            self.units_var.set(units)

    def filter_energy(self, _event=None):
        showing = self.proxy["show"]()
        energy = self.proxy["genre"]()
        factor = 1e-2 if showing == "populations" else 1
        lower = float(self.lower_entry.get()) * factor
        upper = float(self.upper_entry.get()) * factor
        self.tesliper.conformers.trim_to_range(
            energy, minimum=lower, maximum=upper, attribute=showing
        )
        # TODO: turn below into some higher-level method
        for box, kept in zip(
            self.view.trees["main"].boxes.values(),
            self.tesliper.conformers.kept,
        ):
            box.var.set(kept)
        self.event_generate("<<KeptChanged>>")


class FilterRMSD(ttk.Frame):
    def __init__(self, parent, view, proxy, **kwargs):
        super().__init__(parent, **kwargs)
        self.view = view
        self.proxy = proxy  # dict with getter for "genre" combobox

        self.columnconfigure(1, weight=1)

        ttk.Label(self, text="Window size").grid(column=0, row=0)
        ttk.Label(self, text="Threshold").grid(column=0, row=1)
        ttk.Label(self, text="kcal/mol").grid(column=2, row=0)
        ttk.Label(self, text="angstrom").grid(column=2, row=1)
        self.window_size = tk.StringVar(value="5.0")
        self.threshold = tk.StringVar(value="1.0")
        window_size = NumericEntry(
            self,
            textvariable=self.window_size,
            width=4,
            scroll_rate=0.5,
            min_value=0,
        )
        window_size.grid(column=1, row=0, sticky="new")
        threshold = NumericEntry(
            self,
            textvariable=self.threshold,
            width=4,
            scroll_rate=0.1,
            min_value=0,
        )
        threshold.grid(column=1, row=1, sticky="new")
        self.ignore_hydrogens = tk.BooleanVar(value=True)
        ignore_hydrogens = ttk.Checkbutton(
            self, text="Ignore hydrogen atoms", variable=self.ignore_hydrogens
        )
        ignore_hydrogens.grid(column=0, row=3, columnspan=3, sticky="new")

        button = ttk.Button(self, text="Trim similar", command=self._filter)
        button.grid(column=0, row=4, columnspan=3, sticky="nwe")

        # trigger action on Enter key press
        window_size.bind("<Return>", self._filter, "+")
        threshold.bind("<Return>", self._filter, "+")

        self.winfo_toplevel().changer.register(
            [window_size, threshold, ignore_hydrogens, button], "energies"
        )

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    @ThreadedMethod(progbar_msg="Finding similar conformers...")
    def _filter(self, _event=None):
        self.tesliper.conformers.trim_rmsd(
            threshold=float(self.threshold.get()),
            window_size=float(self.window_size.get()),
            energy_genre=self.proxy["genre"](),
            ignore_hydrogen=self.ignore_hydrogens.get(),
        )
        # TODO: turn below into some higher-level method
        for box, kept in zip(
            self.view.trees["main"].boxes.values(),
            self.tesliper.conformers.kept,
        ):
            box.var.set(kept)
        self.event_generate("<<KeptChanged>>")


class FilterEnergies(CollapsiblePane):
    def __init__(self, parent, view, **kwargs):
        super().__init__(parent, text="Energies and structure", **kwargs)
        self.view = view
        root = self.winfo_toplevel()

        ttk.Label(self.content, text="Show:").grid(column=0, row=0, sticky="new")
        self.show_var = tk.StringVar()
        show_values = (
            "Energy /Hartree",
            "Delta /(kcal/mol)",
            "Min. Boltzmann factor",
            "Population /%",
        )
        show_units = ("Hartree", "kcal/mol", "", "%")
        show_id = ("values", "deltas", "min_factors", "populations")
        self.show_ref = {k: v for k, v in zip(show_values, show_id)}
        self.show_units = {k: v for k, v in zip(show_values, show_units)}
        self.show_combo = ttk.Combobox(
            self.content,
            textvariable=self.show_var,
            values=show_values,
            state="readonly",
        )
        self.show_combo.grid(column=1, row=0, sticky="nwe")
        self.show_combo.set("Energy /Hartree")

        # Energy choice
        ttk.Label(self.content, text="Use:").grid(column=0, row=1, sticky="new")
        self.energies_choice = EnergiesChoice(self.content, width=12)
        self.energies_choice.grid(column=1, row=1, sticky="nwe")

        # Temperature value
        t_frame = tk.Frame(self.content)
        t_frame.grid(column=0, row=2, columnspan=2, sticky="we")
        t_frame.columnconfigure((0, 1, 2), weight=1)
        ttk.Label(t_frame, text="Temperature").grid(column=0, row=0, sticky="new")
        self.temperature = NumericEntry(
            t_frame,
            scroll_rate=1,
            decimal_digits=2,
            min_value=0,
            include_min_value=False,
        )
        self.temperature.grid(column=1, row=0)
        ttk.Label(t_frame, text="K").grid(column=2, row=0, sticky="new")
        self.temperature.var.set(self.tesliper.temperature)
        self.temperature.bind("<FocusOut>", self.on_temperature_changed, "+")
        self.temperature.bind("<MouseWheel>", self.on_temperature_changed, "+")
        self.temperature.bind("<Button-4>", self.on_temperature_changed, "+")
        self.temperature.bind("<Button-5>", self.on_temperature_changed, "+")
        root.bind("<<DataExtracted>>", self.update_temperature, "+")
        root.bind("<<Clear>>", self.update_temperature, "+")
        root.changer.register([self.temperature], "tesliper")

        proxy = {
            "genre": self.energies_choice.get_genre,
            "show": lambda: self.show_ref[self.show_var.get()],
            "units": lambda: self.show_units[self.show_var.get()],
        }
        # filter by energy value
        LabelSeparator(self.content, text="Range sieve").grid(
            column=0, row=3, columnspan=2, sticky="nwe"
        )
        self.range = FilterRange(self.content, view=self.view, proxy=proxy)
        self.range.grid(column=0, row=4, columnspan=2, sticky="news")

        # RMSD sieve
        LabelSeparator(self.content, text="RMSD sieve").grid(
            column=0, row=5, columnspan=2, sticky="nwe"
        )
        self.rmsd = FilterRMSD(self.content, view=self.view, proxy=proxy)
        self.rmsd.grid(column=0, row=6, columnspan=2, sticky="news")

        self.show_combo.bind("<<ComboboxSelected>>", self.on_show_selected)
        self.energies_choice.bind("<<ComboboxSelected>>", self.on_energies_selected)
        root.bind("<<DataExtracted>>", self.on_show_selected, "+")
        root.bind("<<KeptChanged>>", self.on_show_selected, "+")
        root.changer.register([self.show_combo, self.energies_choice], "energies")

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    def on_show_selected(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.on_show_selected handler.")
        self.view.refresh(show=self.show_ref[self.show_var.get()])
        self.range.on_show_selected()

    def on_energies_selected(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.on_energies_selected handler.")
        self.range.set_upper_and_lower()

    def update_temperature(self, _event=None):
        self.temperature.var.set(self.tesliper.temperature)

    def on_temperature_changed(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.on_temperature_changed handler.")
        value = self.temperature.get()
        self.tesliper.temperature = float(value)
        # we need it's views-updating side effects
        self.event_generate("<<KeptChanged>>")


OVERVIEW_GENRES = "dip rot vosc vrot losc lrot raman1 roa1 scf zpe ent ten gib".split()


class SelectConformers(CollapsiblePane):
    overview_control_ref = {
        k: v
        for k, v in zip(
            "file en ir vcd uv ecd ram roa incompl term opt imag".split(" "),
            "command gib dip rot vosc vrot raman1 roa1 command "
            "normal_termination optimization_completed freq".split(" "),
        )
    }

    overview_funcs = dict(
        file=lambda *args: True,
        en=lambda *args: "gib" in args[0],
        ir=lambda *args: "dip" in args[0],
        vcd=lambda *args: "rot" in args[0],
        uv=lambda *args: "vosc" in args[0],
        ecd=lambda *args: "vrot" in args[0],
        ram=lambda *args: "raman1" in args[0],
        roa=lambda *args: "roa1" in args[0],
        incompl=lambda *args: not all(g in args[0] for g in args[1]),
        term=lambda *args: not args[0]["normal_termination"],
        opt=lambda *args: "optimization_completed" in args[0]
        and not args[0]["optimization_completed"],
        imag=lambda *args: "freq" in args[0] and any([f < 0 for f in args[0]["freq"]]),
        incons=lambda *args: any(
            g in args[0] and not len(args[0][g]) == mx for g, mx in args[2].items()
        ),
    )

    def __init__(self, parent, view, **kwargs):
        super().__init__(parent, text="Kept conformers", **kwargs)
        self.view = view

        self.widgets = dict()
        self.content.columnconfigure(0, weight=1)
        root = self.winfo_toplevel()
        root.bind("<<KeptChanged>>", self.on_kept_changed, "+")
        root.bind("<<DataExtracted>>", self.on_data_extracted, "+")
        root.bind("<<Clear>>", self.clear, "+")

        count_frame = ttk.Frame(self.content)
        count_frame.grid(column=0, row=0, sticky="news")
        count_frame.columnconfigure((1, 3), weight=1)
        widgets_tuple = namedtuple(
            "widgets", ["label", "count", "slash", "all", "check", "uncheck"]
        )
        for i, (name, key) in enumerate(
            zip(
                "Files Energy IR VCD UV ECD Raman ROA Incompl. Errors "
                "Unopt. Imag.Freq. Incons.".split(),
                "file en ir vcd uv ecd ram roa incompl term opt imag incons".split(),
            )
        ):
            var = tk.IntVar()  # number of conformers selected
            var_all = tk.IntVar()  # number of conformers in total

            label = tk.Label(count_frame, text=name, anchor="w")
            count = tk.Label(count_frame, textvariable=var, bd=0, width=3)
            slash = tk.Label(count_frame, text="/", bd=0)
            all_ = tk.Label(count_frame, textvariable=var_all, bd=0, width=3)
            check_butt = ttk.Button(
                count_frame,
                text="keep",
                width=5,
                command=lambda key=key: self.select(key, keep=True),
            )
            uncheck_butt = ttk.Button(
                count_frame,
                text="trim",
                width=5,
                command=lambda key=key: self.select(key, keep=False),
            )

            count.var = var
            all_.var = var_all

            label.grid(column=0, row=i)
            count.grid(column=1, row=i, sticky="we")
            slash.grid(column=2, row=i)
            all_.grid(column=3, row=i, sticky="we")
            check_butt.grid(column=4, row=i)
            uncheck_butt.grid(column=5, row=i)

            root.changer.register([check_butt, uncheck_butt], "tesliper")

            self.widgets[key] = widgets_tuple(
                label, count, slash, all_, check_butt, uncheck_butt
            )
        separator = LabelSeparator(self.content, text="Auto-trim")
        separator.grid(column=0, row=1, sticky="we")

        keep_unchecked_frame = ttk.Frame(self.content)
        keep_unchecked_frame.grid(column=0, row=2, sticky="nswe")
        self.kept_vars = {
            k: tk.BooleanVar(value=True)
            for k in "error unopt imag stoich incompl incons".split(" ")
        }
        self.kept_buttons = {
            k: ttk.Checkbutton(
                keep_unchecked_frame,
                text=text,
                variable=var,
                command=lambda k=k: self.discard(k),
            )
            for (k, var), text in zip(
                self.kept_vars.items(),
                [
                    "Error termination",
                    "Unoptimised",
                    "Imaginary frequencies",
                    "Non-matching stoichiometry",
                    "Incomplete entries",
                    "Inconsistent data sizes",
                ],
            )
        }
        for n, (key, var) in enumerate(self.kept_vars.items()):
            self.kept_buttons[key].grid(column=0, row=n, sticky="nw")

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    def clear(self, _event=None):
        for items in self.widgets.values():
            items.all.var.set(0)
            items.count.var.set(0)

    def on_data_extracted(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.on_data_extracted handler.")
        # "all" count should only be called after extraction
        self.update_overview_values(untrimmed=True)
        self.on_kept_changed(_event)

    def on_kept_changed(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.on_kept_changed handler.")
        self.discard_not_kept()
        self.update_overview_values()

    def update_overview_values(self, untrimmed=False):
        logger.debug("Called update_overview_values")
        values = {k: 0 for k in self.widgets.keys()}
        try:
            count = [
                [g in conf for g in OVERVIEW_GENRES]
                for conf in self.tesliper.conformers.values()
            ]
            best_match = [g for g, k in zip(OVERVIEW_GENRES, max(count)) if k]
        except ValueError:
            best_match = []
        sizes = {}
        for fname, conf in self.tesliper.conformers.items():
            for genre, value in conf.items():
                if isinstance(value, (np.ndarray, list, tuple)):
                    sizes.setdefault(genre, {})[fname] = len(value)
        maxes = {
            genre: Counter(v for v in values.values()).most_common()[0][0]
            for genre, values in sizes.items()
        }
        conformers = (
            self.tesliper.conformers.values()
            if untrimmed
            else self.tesliper.conformers.kept_values()
        )
        for conf in conformers:
            values["file"] += 1
            values["term"] += not conf["normal_termination"]
            values["incompl"] += not all(g in conf for g in best_match)
            values["opt"] += (
                "optimization_completed" in conf and not conf["optimization_completed"]
            )
            values["imag"] += "freq" in conf and sum(v < 0 for v in conf["freq"]) > 0
            values["en"] += "gib" in conf
            values["ir"] += "dip" in conf
            values["vcd"] += "rot" in conf
            values["uv"] += "vosc" in conf
            values["ecd"] += "vrot" in conf
            values["ram"] += "raman1" in conf
            values["roa"] += "roa1" in conf
            values["incons"] += any(
                g in conf and not len(conf[g]) == mx for g, mx in maxes.items()
            )
        category = "all" if untrimmed else "count"
        for key, items in self.widgets.items():
            getattr(items, category).var.set(values[key])

    def select(self, key, keep):
        confs = self.tesliper.conformers
        condition = self.overview_funcs[key]
        best_match = []
        maxes = {}
        if key == "incompl":
            try:
                count = [
                    [g in conf for g in OVERVIEW_GENRES]
                    for conf in self.tesliper.conformers.values()
                ]
                best_match = [g for g, k in zip(OVERVIEW_GENRES, max(count)) if k]
            except ValueError:
                best_match = []
        elif key == "incons":
            sizes = {}
            for fname, conf in self.tesliper.conformers.items():
                for genre, value in conf.items():
                    if isinstance(value, (np.ndarray, list, tuple)):
                        sizes.setdefault(genre, {})[fname] = len(value)
            maxes = {
                genre: Counter(v for v in values.values()).most_common()[0][0]
                for genre, values in sizes.items()
            }
        for n, conf in enumerate(confs.values()):
            if condition(conf, best_match, maxes):
                self.view.boxes[str(n)].var.set(keep)
        self.event_generate("<<KeptChanged>>")

    def discard(self, key):
        value = self.kept_vars[key].get()
        if key == "incons":
            self.tesliper.conformers.allow_data_inconsistency = not value
        # trigger discard_not_kept(), update_overview_values()
        # and WgtStateChanger.set_states()
        self.event_generate("<<KeptChanged>>")

    def discard_not_kept(self):
        for key, var in self.kept_vars.items():
            if var.get():
                self.kept_funcs[key]()
        for box, kept in zip(self.view.boxes.values(), self.tesliper.conformers.kept):
            box.var.set(kept)

    @property
    def kept_funcs(self):
        return dict(
            error=self.tesliper.conformers.trim_non_normal_termination,
            unopt=self.tesliper.conformers.trim_not_optimized,
            imag=self.tesliper.conformers.trim_imaginary_frequencies,
            stoich=self.tesliper.conformers.trim_non_matching_stoichiometry,
            incompl=self.tesliper.conformers.trim_incomplete,
            incons=self.tesliper.conformers.trim_inconsistent_sizes,
        )


class CalculateSpectra(CollapsiblePane):
    def __init__(self, parent, view, **kwargs):
        super().__init__(parent, text="Calculate Spectra", **kwargs)
        self.view = view
        root = self.winfo_toplevel()
        root.bind("<<Clear>>", self.clear, "+")

        self.content.columnconfigure(0, weight=1)

        # Spectra name
        LabelSeparator(self.content, text="Spectra type").grid(
            column=0, row=0, sticky="we"
        )
        s_name_frame = ttk.Frame(self.content)
        s_name_frame.grid(column=0, row=1, sticky="we")
        s_name_frame.columnconfigure((0, 1), weight=1)
        self.s_name = tk.StringVar()
        self.s_name_radio = {}
        names = "IR UV Raman VCD ECD ROA".split(" ")
        values = "ir uv raman vcd ecd roa".split(" ")
        positions = [(c, r) for c in range(2) for r in range(3)]
        for n, v, (c, r) in zip(names, values, positions):
            b = ttk.Radiobutton(
                s_name_frame,
                text=n,
                variable=self.s_name,
                value=v,
                command=lambda v=v: self.spectra_chosen(v),
                width=6,
            )
            b.configure(state="disabled")
            b.grid(column=c, row=r, padx=5)
            b.bind("<Button-1>", self._stash_current_settings)
            self.s_name_radio[v] = b
            root.changer.register(b, needs_all_genres=[dw.DEFAULT_ACTIVITIES[v]])

        # Settings
        LabelSeparator(self.content, text="Settings").grid(column=0, row=2, sticky="we")
        sett = ttk.Frame(self.content)
        sett.grid(column=0, row=3, sticky="we")
        tk.Grid.columnconfigure(sett, 1, weight=1)
        ttk.Label(sett, text="Fitting").grid(column=0, row=0)
        fit = tk.StringVar()
        self.fitting = ttk.Combobox(sett, textvariable=fit, state="disabled", width=13)
        self.fitting.bind("<<ComboboxSelected>>", self.live_preview_callback)
        self.fitting.var = fit
        self.fitting.grid(column=1, row=0, columnspan=2, sticky="we")
        self.fitting["values"] = ("lorentzian", "gaussian")
        root.changer.register(
            self.fitting, "bars", key=lambda var=self.s_name: bool(var.get())
        )

        scroll_param = {
            "Start": {"scroll_rate": 50},
            "Stop": {"scroll_rate": 50},
            "Step": {"scroll_rate": 1, "min_value": 0, "include_min_value": False},
            "Width": {"scroll_rate": 0.05, "min_value": 0, "include_min_value": False},
            "Offset": {"scroll_rate": 10},
            "Scaling": {"scroll_factor": 1.1},
        }
        for no, name in enumerate("Start Stop Step Width Offset Scaling".split(" ")):
            ttk.Label(sett, text=name).grid(column=0, row=no + 1)
            var = tk.StringVar()
            entry = NumericEntry(
                sett, textvariable=var, width=10, state="disabled", **scroll_param[name]
            )
            entry.bind("<FocusOut>", lambda e: self.live_preview_callback(), "+")
            entry.bind("<MouseWheel>", lambda e: self.live_preview_callback(), "+")
            entry.bind("<Button-4>", lambda e: self.live_preview_callback(), "+")
            entry.bind("<Button-5>", lambda e: self.live_preview_callback(), "+")

            setattr(self, name.lower(), entry)
            entry.var = var
            entry.grid(column=1, row=no + 1, sticky="we", padx=(0, 5))
            unit = tk.StringVar(value="-")
            entry.unit = unit
            label = ttk.Label(sett, textvariable=unit, width=5)
            label.grid(column=2, row=no + 1, sticky="e")
            root.changer.register(
                entry, "bars", key=lambda var=self.s_name: bool(var.get())
            )

        # Calculation Mode
        self.mode = tk.StringVar()
        self.single_radio = ttk.Radiobutton(
            self.content,
            text="Single file",
            variable=self.mode,
            value="single",
            state="disabled",
            command=self.mode_chosen,
        )
        self.single_radio.grid(column=0, row=4, sticky="w")
        self.average_radio = ttk.Radiobutton(
            self.content,
            text="Average by energy",
            variable=self.mode,
            value="average",
            state="disabled",
            command=self.mode_chosen,
        )
        self.average_radio.grid(column=0, row=6, sticky="w")
        self.stack_radio = ttk.Radiobutton(
            self.content,
            text="Stack by overview",
            variable=self.mode,
            value="stack",
            state="disabled",
            command=self.mode_chosen,
        )
        self.stack_radio.grid(column=0, row=8, sticky="w")

        self.single = ConformersChoice(self.content, spectra_var=self.s_name)
        self.single.bind(
            "<<ComboboxSelected>>",
            lambda event: self.live_preview_callback(event, mode="single"),
        )
        self.single.grid(column=0, row=5)
        self.single["values"] = ()
        self.average = EnergiesChoice(self.content)
        self.average.bind(
            "<<ComboboxSelected>>",
            lambda event: self.live_preview_callback(event, mode="average"),
        )
        self.average.grid(column=0, row=7)

        self.stack = ColorsChoice(self.content)
        self.stack.bind("<<ComboboxSelected>>", self.change_colour)
        self.stack.grid(column=0, row=9)
        root.changer.register(
            [self.single_radio, self.single, self.stack_radio, self.stack],
            "bars",
            key=lambda var=self.s_name: bool(var.get()),
        )
        root.changer.register(
            [self.average_radio, self.average],
            ["bars", "energies"],
            key=lambda var=self.s_name: bool(var.get()),
        )
        self.boxes = dict(single=self.single, average=self.average, stack=self.stack)
        self.current_box = None
        for box in self.boxes.values():
            box.grid_remove()

        # Live preview
        # Recalculate
        frame = ttk.Frame(self.content)
        frame.grid(column=0, row=10, sticky="new")
        frame.columnconfigure(0, weight=1)
        var = tk.BooleanVar()
        self.reverse_ax = ttk.Checkbutton(
            frame,
            variable=var,
            text="Reverse x-axis",
            state="disabled",
            command=self.redraw,
        )
        self.reverse_ax.grid(column=0, row=0, sticky="w")
        self.reverse_ax.var = var
        var = tk.BooleanVar(value=True)
        self.show_bars = ttk.Checkbutton(
            frame,
            variable=var,
            text="Show activities",
            state="disabled",
            command=self.redraw,
        )
        self.show_bars.grid(column=0, row=1, sticky="w")
        self.show_bars.var = var
        self.show_bars.previous_value = True
        var = tk.BooleanVar(value=True)
        self.live_prev = ttk.Checkbutton(
            frame, variable=var, text="Live preview", state="disabled"
        )
        self.live_prev.grid(column=0, row=2, sticky="w")
        self.live_prev.var = var
        root.bind("<<KeptChanged>>", self.live_preview_callback, "+")
        # previously labeled 'Recalculate'
        self.recalc_b = ttk.Button(
            frame, text="Redraw", state="disabled", command=self.recalculate_command
        )
        self.recalc_b.grid(column=1, row=2)
        root.changer.register(
            self.reverse_ax, key=lambda var=self.s_name: bool(var.get())
        )
        root.changer.register(
            [self.live_prev, self.recalc_b],
            "bars",
            key=lambda var=self.s_name: bool(var.get()),
        )

        # Experimental spectrum
        LabelSeparator(self.content, text="Experimental spectrum").grid(
            column=0, row=11, sticky="we"
        )
        frame = ttk.Frame(self.content)
        frame.grid(column=0, row=12, sticky="new")
        frame.columnconfigure((0, 1), weight=1)
        var = tk.BooleanVar()
        self.show_exp = ttk.Checkbutton(
            frame,
            variable=var,
            text="Show experimental spectrum",
            state="disabled",
            command=self.redraw,
        )
        self.show_exp.grid(column=0, row=0, columnspan=2, sticky="nw")
        self.show_exp.var = var
        var = tk.BooleanVar(value=True)
        self.allow_double_axis = ttk.Checkbutton(
            frame,
            variable=var,
            text="Allow double y-axis",
            state="disabled",
            command=self.redraw,
        )
        self.allow_double_axis.grid(column=0, row=1, columnspan=2, sticky="nw")
        self.allow_double_axis.var = var
        self.load_exp = ttk.Button(
            frame,
            text="Load from file...",
            state="disabled",
            command=lambda: (self.load_exp_command(), self.live_preview_callback()),
        )
        self.load_exp.grid(column=0, row=2, columnspan=2, sticky="new")
        root.changer.register(
            self.load_exp, key=lambda var=self.s_name: bool(var.get())
        )
        self.auto_scale = ttk.Button(
            frame, text="Auto-scale", state="disabled", command=self.auto_scale_command
        )
        self.auto_scale.grid(column=0, row=3, sticky="new")
        self.auto_shift = ttk.Button(
            frame, text="Auto-shift", state="disabled", command=self.auto_shift_command
        )
        self.auto_shift.grid(column=1, row=3, sticky="new")
        root.changer.register(
            [self.show_exp, self.auto_scale, self.auto_shift, self.allow_double_axis],
            key=lambda wgt=self: wgt.exp_spc is not None,
        )

        self.last_used_settings = {
            name: {
                "offset": 0,
                "scaling": 1,
                "show_bars": True,
                "show_exp": False,
                "reverse_ax": name not in ("uv", "ecd"),
            }
            for name in self.s_name_radio
        }
        self.lastly_drawn_spectra = None

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    @property
    def exp_spc(self):
        try:
            return self.tesliper.experimental[self.s_name.get()]
        except KeyError:
            # no experimental spectrum or no value selected in s_name radio
            return None

    @exp_spc.setter
    def exp_spc(self, value):
        self.tesliper.experimental[self.s_name.get()] = value

    @property
    def draw_params(self):
        spectra_name = self.s_name.get()
        mode = self.mode.get()
        # get value from self.single, self.average or self.stack
        try:
            option = getattr(self, mode).var.get()
        except AttributeError:
            # mode is not chosen
            option = ""
        return {"spectra_name": spectra_name, "mode": mode, "option": option}

    @ThreadedMethod(progbar_msg="Calculating best fit...")
    def _auto_scale(self):
        self.lastly_drawn_spectra.scale_to(self.exp_spc)
        self.scaling.update(self.lastly_drawn_spectra.scaling)
        return self.lastly_drawn_spectra

    def auto_scale_command(self):
        if self.lastly_drawn_spectra is not None:
            self._auto_scale()
            self.draw(**self.draw_params)

    @ThreadedMethod(progbar_msg="Calculating best fit...")
    def _auto_shift(self):
        self.lastly_drawn_spectra.shift_to(self.exp_spc)
        self.offset.update(self.lastly_drawn_spectra.offset)
        return self.lastly_drawn_spectra

    def auto_shift_command(self):
        if self.lastly_drawn_spectra is not None:
            self._auto_shift()
            self.draw(**self.draw_params)

    def load_exp_command(self):
        filename = askopenfilename(
            parent=self,
            title="Select spectrum file.",
            filetypes=[
                ("text files", "*.txt"),
                ("xy files", "*.xy"),
                # ("spc files", "*.spc"),
                # spc not supported yet
                ("all files", "*.*"),
            ],
        )
        if filename:
            logger.debug(f"File: {filename}")
            try:
                self.tesliper.load_experimental(
                    filename, self.draw_params["spectra_name"]
                )
            except ValueError:
                logger.warning(
                    "Experimental spectrum couldn't be loaded. "
                    "Please check if format of your file is supported"
                    " or if file is not corrupted."
                )
            else:
                self.show_exp.var.set(True)
                self.winfo_toplevel().changer.set_states()

    def mode_chosen(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.mode_chosen handler.")
        mode = self.mode.get()
        if self.current_box is not None:
            self.current_box.grid_remove()
        self.current_box = self.boxes[mode]
        self.current_box.grid()
        getattr(self, mode).update_values()  # update linked combobox values
        if mode == "single":
            self.show_bars.config(state="normal")
            self.show_bars.var.set(self.show_bars.previous_value)
        else:
            self.show_bars.config(state="disabled")
            self.show_bars.previous_value = self.show_bars.var.get()
            self.show_bars.var.set(False)
        self.live_preview_callback()

    def spectra_chosen(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.spectra_chosen handler.")
        tslr = self.tesliper
        self.visualize_settings()
        bar = DEFAULT_ACTIVITIES[self.s_name.get()]
        self.single["values"] = [k for k, v in tslr.conformers.items() if bar in v]
        self.winfo_toplevel().changer.set_states()
        if self.mode.get():
            self.live_preview_callback()
        else:
            self.single_radio.invoke()

    def visualize_settings(self):
        spectra_name = self.s_name.get()
        tslr = self.tesliper
        last_used = self.last_used_settings[spectra_name]
        settings = tslr.parameters[spectra_name].copy()
        settings.update(last_used)
        for name, sett in settings.items():
            if name == "fitting":
                try:
                    self.fitting.var.set(settings["fitting"].__name__)
                except AttributeError:
                    self.fitting.var.set(settings["fitting"])
            else:
                entry = getattr(self, name)
                entry.var.set(sett)
                try:
                    entry.unit.set(gw.Spectra._units[spectra_name][name])
                except AttributeError:
                    logger.debug(f"Pass on {name}")
                except KeyError:
                    if name == "offset":
                        entry.unit.set(gw.Spectra._units[spectra_name]["start"])
                    elif name == "scaling":
                        pass
                    else:
                        raise ValueError(f"Invalid setting name: {name}")

    def live_preview_callback(self, _event=None, mode=False):
        if _event is not None:
            logger.debug(f"Event caught by {self}.live_preview_callback handler.")
        spectra_name = self.s_name.get()
        # update appropriate tesliper.parameters with current calculation settings
        if spectra_name:
            # but only if spectra already chosen
            self.tesliper.parameters[spectra_name].update(self.calculation_params)
        # TODO: rewrite condisions' checks more elegantly
        mode_con = self.mode.get() == mode if mode else True
        settings_con = (
            spectra_name not in self.last_used_settings
            or self.current_settings != self.last_used_settings[spectra_name]
        )
        core = any([not self.view.tslr_ax, mode_con, settings_con])
        if core and self.live_prev.var.get():
            logger.debug("Live preview on, spectra will be recalculated.")
            self.recalculate_command()
        else:
            logger.debug("Live preview off.")

    def draw(self, spectra_name, mode, option):
        queue_ = self.winfo_toplevel().thread.queue
        if mode == "single" and self.show_bars.var.get():
            bar_name = dw.DEFAULT_ACTIVITIES[spectra_name]
            with self.tesliper.conformers.trimmed_to([option]):
                bars = self.tesliper[bar_name]
        else:
            bars = None
        self._draw(
            queue_,
            bars=bars,
            colour=option,
            stack=mode == "stack",
            experimental=self.exp_spc if self.show_exp.var.get() else None,
            reverse_ax=self.reverse_ax.var.get(),
            allow_double_axis=self.allow_double_axis.var.get(),
        )

    def _draw(self, queue_, **kwargs):
        try:
            spc = queue_.get(0)  # data put to queue by self._calculate_spectra
            if spc is None:
                # invalid parameters for spectra calculation given, abort
                return
            self.view.draw_spectra(spc, **kwargs)
            self.lastly_drawn_spectra = spc
        except queue.Empty:
            func = functools.update_wrapper(
                functools.partial(self._draw, **kwargs), self._draw
            )
            self.after(20, func, queue_)

    def redraw(self):
        draw_params = self.draw_params
        if draw_params["mode"] == "single" and self.show_bars.var.get():
            bar_name = dw.DEFAULT_ACTIVITIES[draw_params["spectra_name"]]
            with self.tesliper.conformers.trimmed_to([draw_params["option"]]):
                bars = self.tesliper[bar_name]
        else:
            bars = None
        self.view.draw_spectra(
            self.lastly_drawn_spectra,
            bars=bars,
            colour=draw_params["option"],
            stack=draw_params["mode"] == "stack",
            experimental=self.exp_spc if self.show_exp.var.get() else None,
            reverse_ax=self.reverse_ax.var.get(),
            allow_double_axis=self.allow_double_axis.var.get(),
        )

    def change_colour(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.change_colour handler.")
        if self.mode.get() != "stack":
            return
        colour = self.stack.var.get()
        self.view.change_colour(colour)

    def _exec_calculate_spectra(self, spectra_name, conformer=None):
        # should be called in execution thread
        if conformer is not None:
            spc = self.tesliper.calculate_single_spectrum(
                genre=spectra_name, conformer=conformer
            )
        else:
            spectra = self.tesliper.calculate_spectra(spectra_name)
            spc = spectra[spectra_name]  # tslr.calculate_spectra returns dictionary
        offset = self.offset.var.get()
        scaling = self.scaling.var.get()
        if offset:  # if not chosen, ignore
            spc.offset = float(offset)
        if scaling:  # if not chosen, ignore
            spc.scaling = float(scaling)
        return spc

    @ThreadedMethod(progbar_msg="Calculating...")
    def _calculate_spectra(self, spectra_name, option, mode):
        try:
            spc = self._exec_calculate_spectra(
                spectra_name=spectra_name,
                conformer=option if mode == "single" else None,
            )
        except ValueError as error:
            # happens on invalid start/stop/step combination
            logger.warning(error)
            return
        if mode == "average":
            en_name = self.average.get_genre()
            spc = self.tesliper.get_averaged_spectrum(spectra_name, en_name)
        return spc

    @property
    def calculation_params(self):
        d = {
            k: v
            for k, v in self.current_settings.items()
            if k in "start stop step width fitting".split(" ")
        }
        return d

    @property
    def current_settings(self):
        try:
            settings = {
                key: float(getattr(self, key).get())
                for key in "start stop step width offset scaling".split(" ")
            }
            settings.update(
                {
                    key: getattr(self, key).var.get()
                    for key in "reverse_ax show_bars show_exp".split(" ")
                }
            )
            fit = self.fitting.get()
            settings["fitting"] = getattr(dw, fit)
        except ValueError:
            return {}
        return settings

    def _stash_current_settings(self, _event=None):
        spectra_name = self.draw_params["spectra_name"]
        if not spectra_name:
            return  # spectra name radio not selected
        self.last_used_settings[spectra_name] = self.current_settings.copy()

    def recalculate_command(self):
        draw_params = self.draw_params
        spectra_name = draw_params["spectra_name"]
        if not spectra_name:
            logger.debug("Calculation aborted: spectra_name not specified.")
            return
        if not self.current_settings:
            logger.info("Calculation aborted: invalid settings provided.")
            return
        self._stash_current_settings()
        option = draw_params["option"]
        if not option or option.startswith("Choose "):
            logger.info("Calculation aborted: option not chosen.")
            return
        logger.debug(f"Recalculating with {self.current_settings}")
        self._calculate_spectra(**draw_params)
        self.draw(**draw_params)

    def clear(self, _event=None):
        self.s_name.set("")
        self.mode.set("")
        self.show_bars.config(state="disabled")


class ExtractData(ttk.LabelFrame):
    def __init__(self, parent, view, **kwargs):
        self.view = view
        super().__init__(parent, text="Extract data", **kwargs)
        self.columnconfigure(0, weight=1)
        self.b_man_extract = ttk.Button(
            self, text="Choose files", command=self.man_extract
        )
        self.b_man_extract.grid(column=0, row=0, sticky="nwe")
        self.b_auto_extract = ttk.Button(
            self, text="Choose folder", command=self.from_dir
        )
        self.b_auto_extract.grid(column=0, row=1, sticky="nwe")
        self.ignore_unknown = tk.BooleanVar()
        self.check_ignore_unknown = ttk.Checkbutton(
            self, text="Ignore unknown conformers", variable=self.ignore_unknown
        )
        self.check_ignore_unknown.grid(column=0, row=2, sticky="nwe")
        self.winfo_toplevel().changer.register(self.check_ignore_unknown, "tesliper")

    # TODO: add recursive extraction

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    def from_dir(self):
        work_dir = askdirectory()
        if not work_dir:
            return
        wanted = self.tesliper.conformers.keys() if self.ignore_unknown.get() else None
        self.extract(path=work_dir, wanted_files=wanted)

    def man_extract(self):
        files = askopenfilenames(
            filetypes=[
                ("gaussian output", ("*.log", "*.out")),
                ("log files", "*.log"),
                ("out files", "*.out"),
                ("all files", "*.*"),
            ],
            defaultextension=".log",
        )
        if not files:
            return
        paths = [Path(path) for path in files]
        filenames = [path.name for path in paths]
        directory = paths[0].parent
        self.extract(directory, filenames)

    @ThreadedMethod(progbar_msg="Extracting...")
    def extract(self, path, wanted_files=None):
        root = self.winfo_toplevel()
        try:
            for file, data in self.tesliper.extract_iterate(path, wanted_files):
                root.progtext.set(f"Extracting {file}...")
                self.view.insert("", tk.END, text=file)
        except TypeError as err:
            logger.warning("Cannot extract from specified directory: " + err.args[0])
            return
        logger.debug(f"Data extracted from {path}")
        self.event_generate("<<DataExtracted>>")


class ExportData(ttk.LabelFrame):
    def __init__(self, parent, **kwargs):
        # Change label text
        super().__init__(parent, text="Session control", **kwargs)
        root = self.winfo_toplevel()

        tk.Grid.columnconfigure(self, (0, 1), weight=1)
        self.b_save_session = ttk.Button(
            self, text="Save session", command=self.save_session
        )
        self.b_save_session.grid(column=0, row=0, sticky="nwe")
        self.b_load_session = ttk.Button(
            self, text="Load session", command=self.load_session, state="enabled"
        )
        self.b_load_session.grid(column=0, row=1, sticky="nwe")
        self.b_clear_session = ttk.Button(
            self, text="Clear session", command=self.winfo_toplevel().new_session
        )
        self.b_clear_session.grid(column=0, row=2, sticky="nwe")
        root.changer.register([self.b_clear_session, self.b_save_session], "tesliper")

        self.b_text_export = ttk.Button(
            self, text="Export as .txt", command=lambda: self.save(fmt="txt")
        )
        self.b_text_export.grid(column=1, row=0, sticky="nwe")
        self.b_excel_export = ttk.Button(
            self, text="Export as .xlsx", command=lambda: self.save(fmt="xlsx")
        )
        self.b_excel_export.grid(column=1, row=1, sticky="nwe")
        self.b_csv_export = ttk.Button(
            self, text="Export as .csv", command=lambda: self.save(fmt="csv")
        )
        self.b_csv_export.grid(column=1, row=2, sticky="nwe")
        root.changer.register(
            [self.b_text_export, self.b_excel_export, self.b_csv_export], "tesliper"
        )
        self.b_gjf_export = ttk.Button(
            self, text="Create .gjf files", command=self.export_gjf
        )
        self.b_gjf_export.grid(column=1, row=3, sticky="nwe")
        root.changer.register(
            [self.b_gjf_export], needs_any_genre=gw.Geometry.associated_genres
        )

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    def save_session(self):
        file = asksaveasfilename(
            filetypes=[
                ("tesliper", "*.tslr"),
                ("all files", "*.*"),
            ],
            defaultextension=".tslr",
        )
        if not file:
            return
        path = Path(file)
        self.tesliper.output_dir = path.parent
        try:
            self.tesliper.serialize(path.name)
        except FileExistsError:
            if self._should_override([path.name]):
                self.tesliper.serialize(path.name, mode="w")

    def load_session(self):
        file = askopenfilename(
            filetypes=[
                ("tesliper", "*.tslr"),
                ("all files", "*.*"),
            ],
            defaultextension=".tslr",
        )
        if not file:
            return
        root = self.winfo_toplevel()
        root.new_session()
        if not root.tesliper.conformers:
            path = Path(file)
            root.new_tesliper(path)

    def get_save_query(self):
        popup = ExportPopup(self, width="700", height="380")
        query = popup.get_query()
        return query

    @ThreadedMethod(progbar_msg="Saving...")
    def execute_save_command(self, categories, fmt):
        root = self.winfo_toplevel()
        if "spectra" in categories:
            root.progtext.set("Calculating spectra...")
            self.tesliper.spectra = {}
            for spc in categories["spectra"]:
                _ = root.controls.calculate._exec_calculate_spectra(spc.lower())
            root.progtext.set("Saving...")
        if "averaged" in categories:
            root.progtext.set("Averaging spectra...")
            self.tesliper.averaged = {}
            for spc, en in categories["averaged"]:
                averaged = self.tesliper.get_averaged_spectrum(spectrum=spc, energy=en)
                self.tesliper.averaged[(spc, en)] = averaged
            root.progtext.set("Saving...")
        self._exec_save(categories, fmt, mode="x")

    def _exec_save(self, categories, fmt, mode):
        """Executes save command, by selecting writer object appropriate for given fmt
        and invoking its :meth:`.write` method with selected data. If
        :class:`FileExistsError` or :class:`PermissionError` occurres, it gives a popup,
        asking if retry should be attempted. Changed mode to "write" for retry, if
        necessary.
        """
        genres = ["freq", "stoichiometry", "wavelen"]
        genres += categories.get("energies", [])
        genres += categories.get("spectral data", [])
        try:
            idx = genres.index("transitions-all")
        except ValueError:
            transitions_all = False
            logger.debug("transitions-all not requested")
        else:
            _ = genres.pop(idx)
            transitions_all = False
        data = [self.tesliper[g] for g in genres]
        if "spectra" in categories:
            data += [s for s in self.tesliper.spectra.values() if s]
        if "averaged" in categories:
            data += [s for s in self.tesliper.averaged.values() if s]
        retry_mode = None
        try:
            wrt = wr.writer(fmt=fmt, destination=self.tesliper.output_dir, mode=mode)
            wrt.write(data=data)
            if transitions_all:
                wrt.transitions(
                    transitions=self.tesliper["transitions"],
                    wavelengths=self.tesliper["wavelen"],
                    only_highest=False,
                )
        except FileExistsError:
            message = (
                "Some files already exist in this directory. "
                "Would you like to overwrite them?"
            )
            should_override = messagebox.askokcancel(
                parent=self, title="Files already exist!", message=message
            )
            retry_mode = "w" if should_override else None
        except PermissionError as error:
            answer = messagebox.askokcancel(
                "Permission Error", f"Cannot write to file: {error}. Retry?"
            )
            retry_mode = mode if answer else None
        if retry_mode:
            self._exec_save(categories, fmt, mode=retry_mode)

    def save(self, fmt):
        query = self.get_save_query()
        if not query:
            return
        self.tesliper.output_dir = query["dest"]
        logger.info(f"Export requested: {query['query']}; format: {fmt}")
        self.execute_save_command(query["query"], fmt)

    def export_gjf(self):
        query = GjfPopup(self).get_query()
        if not query:
            return
        self.execute_gjf_export(query)

    @ThreadedMethod(progbar_msg="Creating gjf files...")
    def execute_gjf_export(self, query):
        wrt = wr.writer(fmt="gjf", mode="x", **query["init"])
        try:
            wrt.geometry(**query["call"])
        except FileExistsError:
            if self._should_override(["gjf"]):
                wrt.mode = "w"
                wrt.geometry(**query["call"])
        except PermissionError as error:
            messagebox.showwarning(
                "Permission Error", f"Cannot write to file: {error}."
            )



import builtins
from pathlib import Path
from unittest import mock

import pytest

from tesliper.extraction import soxhlet as sx


@pytest.fixture
def files():
    return [Path(p) for p in "a.out b.out b.gjf setup.txt".split()]


@pytest.fixture
def out_files(files):
    return [f for f in files if f.suffix == ".out"]


@pytest.fixture
def lsdir(files):
    with mock.patch.multiple(
        "tesliper.extraction.soxhlet.Path",
        iterdir=mock.Mock(return_value=files),
        is_file=mock.Mock(return_value=True),
    ):
        yield


@pytest.fixture
def sox(lsdir):
    yield sx.Soxhlet()


def test_path_default(sox):
    assert sox.path == Path().resolve()


def test_path_non_existing_path(sox, monkeypatch):
    monkeypatch.setattr(sx.Path, "is_dir", mock.Mock(return_value=False))
    with pytest.raises(FileNotFoundError):
        sox.path = "\\path\\doesnt\\exist"
    sx.Path.is_dir.assert_called()


def test_path_ok(sox, files, monkeypatch):
    monkeypatch.setattr(sx.Path, "is_dir", mock.Mock(return_value=True))
    path = "\\path\\is\\ok"
    sox.path = path
    assert Path(path).resolve() == sox.path
    sx.Path.is_dir.assert_called()
    assert files == sox.all_files


def test_filter_files_no_extension(sox):
    with pytest.raises(ValueError):
        sox.filter_files()


def test_filter_files(sox, out_files):
    assert out_files == sox.filter_files(".out")
    assert [Path("b.gjf")] == sox.filter_files(".gjf")


def test_filter_files_empty(sox):
    assert [] == sox.filter_files(".log")


@pytest.fixture
def log_all_files():
    return [Path(p) for p in "a.log b.log b.gjf".split()]


def test_guess_extension(sox, log_all_files, monkeypatch):
    assert "out" == sox.guess_extension()


@pytest.fixture
def mixed_files():
    return [Path(p) for p in "a.out b.log b.gjf".split()]


def test_guess_extension_mixed(sox, mixed_files, monkeypatch):
    monkeypatch.setattr(sx.Path, "iterdir", mock.Mock(return_value=mixed_files))
    with pytest.raises(ValueError):
        sox.guess_extension()


def test_output_files_mixed(sox, mixed_files, monkeypatch):
    monkeypatch.setattr(sx.Path, "iterdir", mock.Mock(return_value=mixed_files))
    assert sox.output_files == []


def test_guess_extension_missing(sox, monkeypatch):
    monkeypatch.setattr(sx.Path, "iterdir", mock.Mock(return_value=[Path("b.gjf")]))
    with pytest.raises(FileNotFoundError):
        sox.guess_extension()


def test_output_files_missing(sox, monkeypatch):
    monkeypatch.setattr(sx.Path, "iterdir", mock.Mock(return_value=[Path("b.gjf")]))
    assert sox.output_files == []


def test_output_files_missing_explicit_ext(sox, monkeypatch):
    sox.extension = ".out"
    monkeypatch.setattr(sx.Path, "iterdir", mock.Mock(return_value=[Path("b.gjf")]))
    assert sox.output_files == []


def test_output_files(sox, out_files, monkeypatch):
    monkeypatch.setattr(sx.Soxhlet, "guess_extension", mock.Mock(return_value=".out"))
    monkeypatch.setattr(sx.Soxhlet, "filter_files", mock.Mock(return_value=out_files))
    assert out_files == sox.output_files
    sox.guess_extension.assert_called()
    sox.filter_files.assert_called_with(".out")


def test_all_files(sox, files):
    assert files == sox.all_files


@pytest.fixture
def output(monkeypatch, out_files):
    mocked = mock.PropertyMock(return_value=out_files)
    monkeypatch.setattr(sx.Soxhlet, "output_files", mocked)
    return mocked


def test_extract_iter(sox, monkeypatch, output):
    monkeypatch.setattr(sx.Path, "open", mock.mock_open())
    monkeypatch.setattr(sox, "parser", mock.Mock(parse=mock.Mock(return_value={})))
    out = dict(sox.extract_iter())
    assert {"a": {}, "b": {}} == out
    output.assert_called()
    sox.parser.parse.assert_called()
    assert 2 == sox.parser.parse.call_count


@pytest.fixture
def files_tree(tmp_path):
    tree = [
        tmp_path / p
        for p in "a.out b.out b.gjf inner/setup.txt inner/c.out inner/c.gjf".split()
    ]
    for f in tree:
        f.parent.mkdir(parents=True, exist_ok=True)
        with f.open("w") as h:
            h.write("")
    return tree


@pytest.fixture
def rsox(files_tree, tmp_path):
    return sx.Soxhlet(path=tmp_path, recursive=True)


def test_rsox_all_files(rsox, files_tree, tmp_path):
    assert set(rsox.all_files) == set(files_tree)


def test_rsox_output_files(rsox, tmp_path):
    assert set(rsox.output_files) == {
        tmp_path / p for p in "a.out b.out inner/c.out".split()
    }


def test_rsox_no_recursive_output_files(rsox, tmp_path):
    rsox.recursive = False
    assert set(rsox.output_files) == {tmp_path / p for p in "a.out b.out".split()}


def test_parse_one_found_in_path(sox, monkeypatch):
    monkeypatch.setattr(sox.parser, "parse", mock.Mock(side_effect=lambda x: x))
    assert sox.parse_one("setup.txt") == sox.path / "setup.txt"


def test_parse_one_absolute_path(sox, monkeypatch):
    # first is_file called for source appended to sx.path
    monkeypatch.setattr(sx.Path, "is_file", mock.Mock(side_effect=[False, True]))
    monkeypatch.setattr(sox.parser, "parse", mock.Mock(side_effect=lambda x: x))
    assert sox.parse_one("/some/path/setup.txt") == Path("/some/path/setup.txt")


def test_parse_one_file_not_found(sox, monkeypatch):
    monkeypatch.setattr(sx.Path, "is_file", mock.Mock(return_value=False))
    with pytest.raises(FileNotFoundError):
        sox.parse_one("setup.txt")



import pytest
from hypothesis import given
from hypothesis import strategies as st

import tesliper as ts


@given(value=st.floats(max_value=0, allow_nan=False, exclude_max=True))
def test_temperature_below_zero(value):
    tslr = ts.Tesliper()
    with pytest.raises(ValueError):
        tslr.temperature = value


@given(value=st.floats(min_value=0, allow_nan=False, exclude_min=True))
def test_temeprature_clear(value):
    tslr = ts.Tesliper()
    init_value = tslr.temperature
    tslr.temperature = value
    tslr.clear()
    assert tslr.temperature == init_value



import pytest

from tesliper.glassware import DataArray
from tesliper.glassware.arrays import SpectralActivities


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


def get_all_genres_subclassing(cls):
    return {
        g
        for cls in all_subclasses(cls)
        if hasattr(cls, "associated_genres")
        for g in cls.associated_genres
    }


@pytest.fixture
def all_genres():
    return get_all_genres_subclassing(DataArray)


@pytest.fixture
def all_activities():
    return get_all_genres_subclassing(SpectralActivities)


@pytest.fixture(params=get_all_genres_subclassing(DataArray))
def any_genre(request):
    return request.param


@pytest.fixture(params=get_all_genres_subclassing(SpectralActivities))
def any_activity(request):
    return request.param



import distutils.cmd
import subprocess
from distutils.util import convert_path

from setuptools import find_packages, setup

# bypass importing `tesliper`'s content to avoid ImportError
# approach suggested by https://stackoverflow.com/a/24517154/11416569
metadata = {}
md_path = convert_path("tesliper/_metadata.py")
with open(md_path) as md_file:
    exec(md_file.read(), metadata)  # load variables into dict


class BuildBinaryCommand(distutils.cmd.Command):
    description = "build standalone binary file for Windows"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        command = [
            "pyinstaller",
            "--noconfirm",
            "--onefile",
            "--icon",
            ".\\tesliper\\tesliper.ico",
            "--name",
            "Tesliper",
            "--paths",
            ".",
            "--add-data",
            ".\\tesliper\\tesliper.ico;tesliper",
            "--noconsole",
            ".\\bin\\tesliper_gui.py",
        ]
        subprocess.run(command)


with open("README.md", "r") as readme:
    long_desc = readme.read()

extras_require = {
    "gui": ["matplotlib"],
    "test": ["pytest", "hypothesis", "coverage", "pytest-cov"],
    "docs": ["sphinx", "sphinx-rtd-theme"],
    "build": ["pyinstaller", "twine"],
    "dev": ["black", "pre-commit", "flake8", "isort"],
}
extras_require["dev-all"] = sum(extras_require.values(), [])
extras_require["build"] += extras_require["gui"]
extras_require["dev"] += extras_require["gui"] + extras_require["test"]


setup(
    cmdclass={"binary": BuildBinaryCommand},
    name="tesliper",
    version=metadata["__version__"],
    description=(
        "a package for batch processing of spectra-related Gaussian output files"
    ),
    long_description=long_desc,
    long_description_content_type="text/markdown",
    author=metadata["__author__"],
    author_email="wieclawmm@gmail.com",
    url="https://github.com/mishioo/tesliper",
    packages=find_packages(),
    install_requires=["numpy", "openpyxl"],
    extras_require=extras_require,
    python_requires=">=3.6",
    scripts=["bin/tesliper_gui.py"],
    entry_points={"console_scripts": ["tesliper-gui=tesliper.gui:run"]},
    package_data={"tesliper": ["tesliper.ico"]},
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Development Status :: 4 - Beta",
    ],
    keywords=[
        "chemistry",
        "chemical computing",
        "optical spectroscopy",
        "spectral simulations",
        "spectroscopy",
        "Gaussian",
        "workflow automation",
        "batch processing",
    ],
)



from tkinter import ttk


class LabelSeparator(ttk.Frame):
    """A horizontal separator with centered text over it."""

    # based on https://stackoverflow.com/q/38396900/11416569
    def __init__(self, parent, text="", **kwargs):
        super().__init__(parent, **kwargs)
        self.columnconfigure(0, weight=1)

        self.separator = ttk.Separator(self, orient="horizontal")
        self.separator.grid(row=0, column=0, sticky="ew")

        self.label = ttk.Label(self, text=text)
        self.label.grid(row=0, column=0)



# IMPORTS
import logging as lgg
import re
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import messagebox

# LOGGER
from tkinter.filedialog import askdirectory

from tesliper import datawork as dw
from tesliper.glassware import (
    ElectronicActivities,
    ElectronicData,
    ScatteringActivities,
    ScatteringData,
    SpectralActivities,
    VibrationalActivities,
    VibrationalData,
)

from . import CollapsiblePane
from .choices import GeometriesChoice
from .helpers import WgtStateChanger
from .hinted_entry import HintedEntry
from .label_separator import LabelSeparator
from .numeric_entry import IntegerEntry

logger = lgg.getLogger(__name__)


def not_implemented_popup():
    messagebox.showinfo(
        "Sorry!", "We are sorry, but this function is not implemented yet."
    )


# CLASSES
class Popup(tk.Toplevel):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.grab_set()
        try:
            width, height = kwargs["width"], kwargs["height"]
            self.set_geometry(width, height)
        except KeyError:
            pass

    def set_geometry(self, width, height):
        x = self.master.winfo_pointerx()
        y = self.master.winfo_pointery()
        geometry = "{}x{}{:+n}{:+n}".format(width, height, x, y)
        self.geometry(geometry)


class EnergiesDetails(ttk.Frame):
    def __init__(self, master, **kwargs):
        # TODO: add choice for corrections, stoichiometry, and imaginary
        style = kwargs.pop("style", "active.TFrame")
        super().__init__(master, style=style, **kwargs)
        self.genres = "ten ent gib scf zpe".split(" ")
        self.labels = "Thermal Enthalpy Gibbs SCF Zero-Point".split(" ")
        self.vars = [
            tk.BooleanVar(value=master.tesliper.conformers.has_genre(g))
            for g in self.genres
        ]
        self.checks = [
            ttk.Checkbutton(self, text=label, variable=var, style="active.TCheckbutton")
            for label, var in zip(self.labels, self.vars)
        ]
        for num, check in enumerate(self.checks):
            check.grid(column=0, row=num, padx=5, sticky="nws")
        for check, genre in zip(self.checks, self.genres):
            master.changer.register([check], needs_all_genres=[genre])
        self.rowconfigure((0, 1, 2, 3, 4), weight=1)

    def get_query(self):
        return [g for g, v in zip(self.genres, self.vars) if v.get()]


class SpectralDataDetails(ttk.Frame):
    def __init__(self, master, **kwargs):
        style = kwargs.pop("style", "active.TFrame")
        super().__init__(master, style=style, **kwargs)
        self.vars = {}
        self.columnconfigure(0, weight=1)
        defaults = set("dip rot vdip vrot transitions raman1 roa1".split())
        cols = 3
        classes = [
            VibrationalActivities,
            VibrationalData,
            ElectronicActivities,
            ElectronicData,
            ScatteringActivities,
            ScatteringData,
        ]
        frames = dict()
        transitions_names = {
            "transitions": "Highest contrib. transitions",
            "transitions-all": "All transitions",
        }
        for num, class_ in enumerate(classes):
            name = re.findall("[A-Z][^A-Z]*", class_.__name__)[0] + " Data"
            if name not in frames:
                sep = LabelSeparator(self, text=name, style="active.TFrame")
                sep.label.configure(style="active.TLabel")
                sep.grid(column=0, row=num * 2, sticky="news")
                frame = ttk.Frame(self, style="active.TFrame")
                frame.grid(column=0, row=num * 2 + 1, sticky="news")
                frame.columnconfigure(tuple(range(cols)), weight=1)
                last_idx = 0
            else:
                frame, last_idx = frames[name]
            associated_genres = class_.associated_genres
            if class_ is ElectronicActivities:
                # add transitions
                associated_genres = associated_genres + tuple(transitions_names)
            for idx_, genre in enumerate(associated_genres):
                if genre in transitions_names:
                    full_name = transitions_names[genre]
                else:
                    try:
                        full_name = class_._full_name_ref[genre]
                    except KeyError:
                        full_name = genre

                idx = last_idx + idx_
                val = genre in defaults and master.tesliper.conformers.has_genre(genre)
                var = tk.BooleanVar(value=val)
                self.vars[genre] = var
                cb = ttk.Checkbutton(
                    frame,
                    text=full_name,
                    variable=var,
                    style="active.TCheckbutton",
                    width=30,
                )
                cb.grid(column=idx % cols, row=idx // cols, padx=(5, 0), sticky="news")
                genre_needed = "transitions" if genre == "transitions-all" else genre
                master.changer.register([cb], needs_all_genres=[genre_needed])
            try:
                frames[name] = frame, idx + 1
            except NameError:
                frames[name] = frame, last_idx
        for frame, idx in frames.values():
            self.rowconfigure(frame.grid_info()["row"], weight=idx)

    def get_query(self):
        return [g for g, v in self.vars.items() if v.get()]


class SpectraDetails(ttk.Frame):
    def __init__(self, master, **kwargs):
        style = kwargs.pop("style", "active.TFrame")
        super().__init__(master, style=style, **kwargs)
        self.vars = {}
        spectra = "IR VCD UV ECD Raman ROA".split(" ")
        self.rowconfigure((0, 1, 2, 3, 4, 5), weight=1)
        calculations = master.master.winfo_toplevel().controls.calculate
        for idx, spc in enumerate(spectra):
            spectra_name = spc.lower()
            default_params = master.tesliper.standard_parameters[spectra_name]
            last_used = {
                k: calculations.last_used_settings[spectra_name][k]
                for k in "start stop step width fitting".split(" ")
                if k in calculations.last_used_settings[spectra_name]
            }
            default = not last_used or last_used == default_params

            act_genre = dw.DEFAULT_ACTIVITIES[spectra_name]
            var = tk.BooleanVar(value=master.tesliper.conformers.has_genre(act_genre))
            self.vars[spc] = var
            cb = ttk.Checkbutton(
                self, text=spc, variable=var, style="active.TCheckbutton"
            )
            cb.grid(column=0, row=idx, padx=(5, 0), sticky="nws")
            master.changer.register([cb], needs_all_genres=[act_genre])
            text = "[user parameters]" if not default else "[default parameters]"
            label = ttk.Label(
                self, text=text if var.get() else "", style="active.TLabel"
            )
            label.grid(column=1, row=idx, padx=(5, 0), sticky="nws")
            cb.configure(
                command=lambda t=text, w=label, v=var: w.configure(
                    text=t if v.get() else ""
                )
            )

    def get_query(self):
        return [g for g, v in self.vars.items() if v.get()]


class AveragedDetails(ttk.Frame):
    def __init__(self, master, **kwargs):
        style = kwargs.pop("style", "active.TFrame")
        super().__init__(master, style=style, **kwargs)
        self.columnconfigure((1, 2, 3, 4, 5), weight=1)
        self.rowconfigure((1, 2, 3, 4, 5, 6), weight=1)
        self.vars = {}
        self.checks = {}
        self.alls = {}
        spectra = "IR VCD UV ECD Raman ROA".split(" ")
        energy_names = "Thermal Enthalpy Gibbs SCF Zero-Point".split(" ")
        energy_genres = "ten ent gib scf zpe".split(" ")

        # for checking default values
        has_it = {en: master.tesliper.conformers.has_genre(en) for en in energy_genres}
        for spectra_name in (s.lower() for s in spectra):
            act_genre = dw.DEFAULT_ACTIVITIES[spectra_name]
            has_it[spectra_name] = master.tesliper.conformers.has_genre(act_genre)

        # labels and checkboxes
        for col, energy in enumerate(energy_names):
            ttk.Label(
                self, text=energy, style="active.TLabel", width=9, anchor="center"
            ).grid(column=1 + col, row=0, pady=(3, 0), sticky="news")
        for row, spc in enumerate(spectra):
            ttk.Label(self, text=spc, style="active.TLabel").grid(
                column=0, row=1 + row, pady=(3, 0), sticky="news"
            )
            for col, en in enumerate(energy_genres):
                label = (spc.lower(), en)
                var = tk.BooleanVar(value=has_it[spc.lower()] and has_it[en])
                cb = ttk.Checkbutton(
                    self,
                    style="checkbox.active.TCheckbutton",
                    variable=var,
                    command=lambda label=label: self.single_clicked(label),
                )
                cb.grid(column=1 + col, row=1 + row)
                act_genre = dw.DEFAULT_ACTIVITIES[spc.lower()]
                master.changer.register([cb], needs_all_genres=[act_genre, en])
                self.vars[label] = var
                self.checks[label] = cb

        # buttons for select/disselect all in row/column
        for row, spc in enumerate(s.lower() for s in spectra):
            cb = ttk.Checkbutton(self, style="checkbox.active.TCheckbutton")
            cb.configure(command=lambda cb=cb, idx=spc: self.all_clicked(cb, idx))
            cb.grid(column=6, row=1 + row)
            act_genre = dw.DEFAULT_ACTIVITIES[spc]
            master.changer.register([cb], needs_all_genres=[act_genre])
            self.alls[spc] = cb
        for col, en in enumerate(energy_genres):
            cb = ttk.Checkbutton(self, style="checkbox.active.TCheckbutton")
            cb.configure(command=lambda cb=cb, idx=en: self.all_clicked(cb, idx))
            cb.grid(column=1 + col, row=7)
            master.changer.register([cb], needs_all_genres=[en])
            self.alls[en] = cb
        ttk.Label(
            self, text="All", style="active.TLabel", width=9, anchor="center"
        ).grid(column=col + 2, row=0, pady=(3, 0), sticky="news")
        ttk.Label(self, text="All", style="active.TLabel").grid(
            column=0, row=row + 2, pady=(3, 0), sticky="news"
        )

    def all_clicked(self, cb, idx):
        for key, var in self.vars.items():
            if idx in key:
                theother = tuple(set(key) - set([idx]))[0]
                var.set(
                    # never set disabled checkbox as "selected"
                    str(self.checks[key]["state"]) != "disabled"
                    and cb.instate(["selected"])
                )
                self.set_all_box(theother)

    def set_all_box(self, idx):
        vals = [
            v.get()
            for k, v in self.vars.items()
            # ignore disabled checkboxes
            if idx in k and str(self.checks[k]["state"]) != "disabled"
        ]
        all_same = all(vals[0] == val for val in vals[1:])
        if vals and all(vals):
            self.alls[idx].state(["!alternate", "selected"])
        elif not all_same:
            self.alls[idx].state(["alternate", "!selected"])
        else:
            self.alls[idx].state(["!alternate", "!selected"])

    def update_all_boxes(self):
        # should be called after updating states of checkboxes
        for idx in self.alls:
            self.set_all_box(idx)

    def single_clicked(self, label):
        for idx in label:
            self.set_all_box(idx)

    def get_query(self):
        return [(s, e) for (s, e), v in self.vars.items() if v.get()]


class ExportPopup(Popup):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.title("Export...")
        root = master.winfo_toplevel()
        self.tesliper = root.tesliper
        self.changer = WgtStateChanger(root)
        self.rowconfigure((1, 2, 3, 4), weight=1)
        self.columnconfigure(1, weight=1)

        path_frame = ttk.Frame(self)
        path_frame.grid(column=0, row=0, columnspan=2, sticky="new")
        path_frame.columnconfigure(1, weight=1)
        ttk.Label(path_frame, text="Path").grid(
            column=0, row=0, padx=5, pady=3, sticky="new"
        )
        self.path = tk.StringVar()
        self.path.set(str(self.tesliper.output_dir))
        self.path_entry = ttk.Entry(
            path_frame, textvariable=self.path, state="readonly"
        )
        self.path_entry.grid(column=1, row=0, sticky="ew")
        self.browse = ttk.Button(path_frame, text="Browse", command=self._browse)
        self.browse.grid(column=2, row=0, sticky="we", padx=5)

        self.labels = ["Energies", "Spectral data", "Spectra", "Averaged"]
        self.vars = [tk.BooleanVar() for _ in self.labels]
        self.checks = []
        self.details = []

        details_frames_class = {
            "Energies": EnergiesDetails,
            "Spectral data": SpectralDataDetails,
            "Spectra": SpectraDetails,
            "Averaged": AveragedDetails,
        }
        style = ttk.Style()
        active_color = "ghost white"
        style.configure("active.TFrame", background=active_color)
        style.configure("active.TCheckbutton", background=active_color)
        style.configure("active.TLabel", background=active_color)
        style.layout(
            "checkbox.active.TCheckbutton",
            [("Checkbutton.indicator", {"side": "left", "sticky": ""})],
        )
        for n, (label, var) in enumerate(zip(self.labels, self.vars)):
            tab = ttk.Frame(self, style="TabLike.TFrame")
            tab.grid(column=0, row=n + 1, sticky="news")
            check = ttk.Checkbutton(tab, text=label, variable=var)
            check.grid(column=0, row=0, pady=10, padx=5, sticky="w")
            self.checks.append(check)
            details = details_frames_class[label](self)
            details.grid(column=1, row=1, rowspan=4, sticky="news")
            details.grid_remove()
            kwargs = {"tab": tab, "checkbox": check, "details": details}
            tab.bind("<Enter>", lambda _e, kw=kwargs: self.on_tab_enter(**kw))
            tab.bind("<Leave>", lambda _e, kw=kwargs: self.on_tab_leave(**kw))
            details.bind("<Leave>", lambda _e, kw=kwargs: self.on_tab_leave(**kw))
            self.details.append(details)

        self.details_frame = ttk.Frame(self)
        self.details_frame.grid(column=1, row=1, rowspan=4, sticky="news")
        self.details_frame.columnconfigure(0, weight=1)
        self.details_frame.rowconfigure(0, weight=1)
        ttk.Label(
            self.details_frame, text="Hover over the left side\nto see details."
        ).grid(column=0, row=0)

        self.changer.register(self.checks[0], dependencies="energies")
        self.changer.register(self.checks[1:], dependencies="bars")
        self.changer.set_states()
        self.details[-1].update_all_boxes()
        self.protocol("WM_DELETE_WINDOW", self.cancel_command)
        buttons_frame = ttk.Frame(self)
        buttons_frame.grid(column=0, row=5, pady=2, columnspan=2, sticky="se")
        b_cancel = ttk.Button(buttons_frame, text="Cancel", command=self.cancel_command)
        b_cancel.grid(column=0, row=0, sticky="se")
        b_ok = ttk.Button(buttons_frame, text="OK", command=self.ok_command)
        b_ok.grid(column=1, row=0, padx=5, sticky="se")
        self.query = {}

    def on_tab_enter(self, tab, checkbox, details):
        if str(checkbox["state"]) == tk.DISABLED:
            return
        tab.configure(style="active.TFrame")
        checkbox.configure(style="active.TCheckbutton")
        self.details_frame.grid_remove()
        details.grid()

    def on_tab_leave(self, tab, checkbox, details):
        under_mouse = self.winfo_containing(*self.winfo_pointerxy())
        logger.debug(f"Currently under pointer: {under_mouse}")
        if str(under_mouse).startswith((str(tab), str(details))):
            return
        tab.configure(style="TFrame")
        checkbox.configure(style="TCheckbutton")
        details.grid_remove()
        self.details_frame.grid()

    def _browse(self):
        directory = askdirectory()
        if not directory:
            return
        self.path.set(directory)

    def ok_command(self):
        vals = [v.get() for v in self.vars]
        logger.debug(vals)
        if any(vals):
            self.destroy()
        else:
            messagebox.showinfo(
                "Nothing choosen!", "You must chose what you want to extract."
            )
            self.focus_set()

    def cancel_command(self):
        self.query = None
        self.destroy()

    def get_query(self):
        self.wait_window()
        if self.query is None:
            return {}
        self.query["dest"] = self.path.get()
        self.query["query"] = {
            thing.lower(): details.get_query()
            for thing, var, details in zip(self.labels, self.vars, self.details)
            if var.get()
        }
        logger.debug(self.query)
        return self.query


class LinkZero(ttk.Frame):
    def __init__(self, master, **kwargs):
        super().__init__(master, **kwargs)
        tip_text = (
            "Tip: you may use ${conf} placeholder in path specification to get "
            "a conformer's name or ${num} to get a sequential number."
        )
        self.tip = ttk.Label(
            self, text=tip_text, wraplength=400, justify="left", foreground="gray"
        )

        self.descriptions = {
            "Mem": "amount of dynamic memory used",
            "Chk": "path to the checkpoint file",
            "OldChk": "path to the previous checkpoint file",
            "SChk": "path to which checkpoint file is copied on job step start",
            "RWF": "path to the read-write file",
            "OldMatrix": "path to the unformatted binary matrix element file",
            "OldRawMatrix": "path to the raw binary matrix element file",
            "Int": "path to the two-electron integral file(s)",
            "D2E": "path to the two-electron integral derivative file(s)",
            "KJob": "link number and, optionally, space-separated number",
            "Save": "empty for false, anything else for true",
            "ErrorSave": "aka 'NoSave', empty for false, anything else for true",
            "Subst": "link number and space-separated file path",
        }
        self.command = tk.StringVar()
        self.command_checkbox = ttk.Combobox(
            self,
            textvariable=self.command,
            state="readonly",
            values=list(self.descriptions.keys()),
            width=13,
        )
        self.command_checkbox.bind("<<ComboboxSelected>>", self.combobox_selected)
        self.value = HintedEntry(
            self, hint="← select a link0 command", state="readonly"
        )
        self.value.bind("<Return>", self.add)
        self.add_button = ttk.Button(
            self, text="+", command=self.add, state="disabled", width=1
        )
        self.items_frame = ttk.Frame(self)
        self.items = {}

        self.columnconfigure(1, weight=1)
        self.tip.grid(column=0, row=0, columnspan=3, sticky="ew")
        self.command_checkbox.grid(column=0, row=1, sticky="ew")
        self.value.grid(column=1, row=1, sticky="ew")
        self.add_button.grid(column=2, row=1, sticky="ew")
        self.items_frame.grid(column=0, row=2, columnspan=3, sticky="news")
        self.items_frame.columnconfigure(1, weight=1)

    def edit(self, item, value):
        self.items[item]["value"].configure(text=value)

    def add(self, _event=None):
        item = self.command.get()
        value = self.value.get()
        if item in self.items:
            return self.edit(item, value)
        idx = len(self.items)
        self.items[item] = {
            "command": ttk.Label(self.items_frame, text=item, width=13),
            "value": ttk.Label(self.items_frame, text=value, anchor="e"),
            "button": ttk.Button(
                self.items_frame,
                text="-",
                command=lambda i=item: self.remove(i),
                width=1,
            ),
        }
        self.items[item]["command"].grid(column=0, row=idx, sticky="ew")
        self.items[item]["value"].grid(column=1, row=idx, sticky="ew")
        self.items[item]["button"].grid(column=2, row=idx, sticky="ew")

    def remove(self, item):
        for widget in self.items[item].values():
            widget.grid_forget()
            widget.destroy()
        del self.items[item]

    def combobox_selected(self, _event=None):
        self.add_button.configure(state="normal")
        self.value.configure(state="normal")
        item = self.command.get()
        if item in self.items:
            self.value.set(self.items[item]["value"]["text"])
        else:
            self.value.configure(hint=self.descriptions[item])
            self.value.set("")

    def get_query(self):
        return {key: value["value"].cget("text") for key, value in self.items.items()}


class GjfPopup(Popup):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.title("Create .gjf files...")
        root = master.winfo_toplevel()
        self.tesliper = root.tesliper
        self.changer = WgtStateChanger(root)
        self.query = {}

        self.columnconfigure(1, weight=1)
        path_frame = ttk.Frame(self)
        path_frame.grid(column=0, row=0, columnspan=2, sticky="new")
        path_frame.columnconfigure(1, weight=1)
        ttk.Label(path_frame, text="Path").grid(
            column=0, row=0, padx=5, pady=3, sticky="new"
        )
        self.path = tk.StringVar()
        self.path.set(str(self.tesliper.output_dir))
        self.path_entry = ttk.Entry(
            path_frame, textvariable=self.path, state="readonly"
        )
        self.path_entry.grid(column=1, row=0, sticky="ew")
        self.browse = ttk.Button(path_frame, text="Browse", command=self._browse)
        self.browse.grid(column=2, row=0, sticky="we", padx=5)

        frame = ttk.Frame(self)
        frame.grid(column=0, row=1, columnspan=2, pady=3, sticky="new")
        # geometry genre selection
        ttk.Label(frame, text="Geometry type").grid(
            column=0, row=0, padx=5, sticky="new"
        )
        self.geom_combobox = GeometriesChoice(frame)
        self.geom_combobox.grid(column=1, row=0, padx=(0, 5), sticky="new")
        self.geom_combobox.update_values()
        # charge and multiplicity
        ttk.Label(frame, text="Multiplicity").grid(
            column=2, row=0, padx=(0, 5), sticky="new"
        )
        self.multiplicity = tk.StringVar(value="1")
        self.multiplicity_entry = IntegerEntry(
            frame, textvariable=self.multiplicity, width=4
        )
        self.multiplicity_entry.grid(column=3, row=0, padx=(0, 5), sticky="new")
        ttk.Label(frame, text="Charge").grid(column=4, row=0, padx=(0, 5), sticky="new")
        self.charge = tk.StringVar(value="0")
        self.charge_entry = IntegerEntry(frame, textvariable=self.charge, width=4)
        self.charge_entry.grid(column=5, row=0, padx=(0, 5), sticky="new")

        # job route entry
        ttk.Label(self, text="Route").grid(column=0, row=2, padx=5, sticky="new")
        self.job = HintedEntry(self, hint="Calculations specification")
        self.job.grid(column=1, row=2, padx=(0, 5), sticky="new")

        # comment / job description
        ttk.Label(self, text="Comment").grid(column=0, row=3, padx=5, sticky="new")
        self.comment = HintedEntry(self, hint="Short description, optional")
        self.comment.grid(column=1, row=3, padx=(0, 5), sticky="new")

        # link0 commands
        pane = CollapsiblePane(self, text="Link0 commands", collapsed=True)
        pane.grid(column=0, row=4, columnspan=2, padx=5, pady=3, sticky="new")
        pane.content.columnconfigure(0, weight=1)
        pane.content.rowconfigure(0, weight=1)
        self.link_zero = LinkZero(pane.content)
        self.link_zero.grid(column=0, row=0, sticky="news")

        # after-geometry specifications
        pane = CollapsiblePane(
            self, text="Post-geometry specifications", collapsed=True
        )
        pane.grid(column=0, row=5, columnspan=2, padx=5, pady=3, sticky="news")
        pane.content.columnconfigure(0, weight=1)
        pane.content.rowconfigure(0, weight=1)
        # post_spec width controlled by grid manager, hence 0 below
        self.post_spec = tk.Text(pane.content, height=4, width=0)
        self.post_spec.grid(column=0, row=0, sticky="news")
        self.rowconfigure(5, weight=1)

        # ok / cancel buttons
        buttons_frame = ttk.Frame(self)
        buttons_frame.grid(column=0, row=6, pady=2, columnspan=2, sticky="se")
        b_cancel = ttk.Button(buttons_frame, text="Cancel", command=self.cancel_command)
        b_cancel.grid(column=0, row=0, sticky="se")
        b_ok = ttk.Button(buttons_frame, text="OK", command=self.ok_command)
        b_ok.grid(column=1, row=0, padx=5, sticky="se")

    def _browse(self):
        directory = askdirectory()
        if not directory:
            return
        self.path.set(directory)

    def ok_command(self):
        self.query["init"] = {
            "destination": self.path.get(),
            "link0": self.link_zero.get_query(),
            "route": self.job.get(),
            "comment": self.comment.get() or "No information provided.",
            "post_spec": self.post_spec.get("1.0", "end"),  # Tk.Text instance
        }
        self.query["call"] = {
            "geometry": self.tesliper[self.geom_combobox.get_genre()],
            "charge": int(self.charge.get()),
            "multiplicity": int(self.multiplicity.get()),
        }
        self.destroy()

    def cancel_command(self):
        self.query = {}
        self.destroy()

    def get_query(self):
        self.wait_window()
        logger.debug(f"Request: {self.query}")
        return self.query



import logging
import tkinter as tk
import tkinter.ttk as ttk
from abc import ABC, abstractmethod

from tesliper import datawork as dw

# LOGGER
logger = logging.getLogger(__name__)


# CLASSES
class AutoComboboxBase(ttk.Combobox, ABC):
    """Combobox implementing functionality for automatically updating list of available
    values."""

    def __init__(self, parent, **kwargs):
        self.var = tk.StringVar()
        kwargs["textvariable"] = self.var
        kwargs["state"] = "readonly"
        super().__init__(parent, **kwargs)
        root = self.winfo_toplevel()
        root.bind("<<KeptChanged>>", self.update_values, "+")
        root.bind("<<DataExtracted>>", self.update_values, "+")
        root.bind("<<Clear>>", self.update_values, "+")

    @abstractmethod
    def get_available_values(self):
        raise NotImplementedError

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    def update_values(self, _event=None):
        """Update displayed values to reflect currently available energy genres.
        If previously chosen genre is no longer available, change it."""
        if _event is not None:
            logger.debug(f"Event caught by {self}.update_values handler.")
        current = self.var.get()
        available = self.get_available_values()
        self["values"] = available
        logger.debug(f"Updated {self} values with {available}.")
        if available and current not in available:
            self.var.set(available[0])
            logger.info(
                f"Option '{current}' is no longer available, "
                f"changed to {available[0]}."
            )
        elif not available:
            self.var.set("Not available.")
            logger.info("No values available, removed selection.")


class GenresChoice(AutoComboboxBase):
    """Combobox that enables choice of type of energy."""

    def __init__(self, parent, genres, names=None, **kwargs):
        super().__init__(parent, **kwargs)
        if names is not None and len(names) != len(genres):
            raise ValueError(
                "One name is needed for each genre. "
                f"Got {len(names)} names and {len(genres)} genres."
            )
        if names is None:
            names = genres
        self._names_ref = {k: v for k, v in zip(names, genres)}
        self._genres_ref = {v: k for k, v in self._names_ref.items()}

    def get_genre(self):
        """Convenience method for getting genre of the energy type chosen."""
        return self._names_ref[self.var.get()]

    def get_available_values(self):
        available_genres = [
            genre
            for genre in self._genres_ref
            if self.tesliper.conformers.has_genre(genre)
        ]
        available = tuple(self._genres_ref[genre] for genre in available_genres)
        return available


class EnergiesChoice(GenresChoice):
    def __init__(self, parent, **kwargs):
        super().__init__(
            parent=parent,
            genres="ten ent gib scf zpe".split(),
            names="Thermal Enthalpy Gibbs SCF Zero-Point".split(),
            **kwargs,
        )


class GeometriesChoice(GenresChoice):
    def __init__(self, parent, **kwargs):
        super().__init__(
            parent=parent,
            genres=["last_read_geom", "input_geom", "optimized_geom"],
            names=["Last read", "Input", "Optimized"],
            **kwargs,
        )


class ConformersChoice(AutoComboboxBase):
    """Combobox that enables choice of conformer for spectra calculation."""

    def __init__(self, parent, spectra_var, **kwargs):
        super().__init__(parent, **kwargs)
        self.spectra_var = spectra_var

    def get_available_values(self):
        """Returns filenames of conformers having data for chosen spectra."""
        try:
            activities_genre = dw.DEFAULT_ACTIVITIES[self.spectra_var.get()]
        except KeyError:
            return []
        available = self.tesliper[activities_genre].filenames.tolist()
        return available


class ColorsChoice(AutoComboboxBase):
    def get_available_values(self):
        return (
            "viridis plasma spring summer autumn winter copper rainbow "
            "turbo gnuplot Blues Reds Greens Greys ".split()
        )



import logging as lgg
import tkinter.ttk as ttk

import numpy as np
from matplotlib import cm
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

logger = lgg.getLogger(__name__)


def align_yaxis(axes):
    """Align zeros of the axes."""
    # credits: https://stackoverflow.com/a/59564220/11416569
    y_lims = np.array([ax.get_ylim() for ax in axes])

    # force 0 to appear on all axes, comment if don't need
    y_lims[:, 0] = y_lims[:, 0].clip(None, 0)
    y_lims[:, 1] = y_lims[:, 1].clip(0, None)

    # normalize all axes
    y_mags = (y_lims[:, 1] - y_lims[:, 0]).reshape(len(y_lims), 1)
    y_lims_normalized = y_lims / y_mags

    # find combined range
    y_new_lims_normalized = np.array(
        [np.min(y_lims_normalized), np.max(y_lims_normalized)]
    )

    # denormalize combined range to get new axes
    new_lims = y_new_lims_normalized * y_mags
    for i, ax in enumerate(axes):
        ax.set_ylim(new_lims[i])


class SpectraView(ttk.Frame):
    """Frame embedding matplotlib's canvas for drawing spectra."""

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.figure = Figure()
        # ensure proper plot resizing
        self.bind("<Configure>", lambda event: self.figure.tight_layout())
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.get_tk_widget().grid(column=0, row=0, sticky="nwse")
        self.tslr_ax = None
        self.bars_ax = None
        self.exp_ax = None
        self.winfo_toplevel().bind("<<Clear>>", self.clear, "+")
        # TODO: add save/save img buttons

    def clear(self, _event=None):
        self.new_plot()
        self.canvas.draw()

    def new_plot(self):
        if self.tslr_ax:
            self.figure.delaxes(self.tslr_ax)
            self.tslr_ax = None
        if self.bars_ax:
            self.figure.delaxes(self.bars_ax)
            self.bars_ax = None
        if self.exp_ax:
            self.figure.delaxes(self.exp_ax)
            self.exp_ax = None

    def draw_spectra(
        self,
        spc,
        bars=None,
        colour=None,
        width=0.5,
        stack=False,
        experimental=None,
        reverse_ax=False,
        allow_double_axis=True,
    ):
        # TO DO: correct spectra drawing when offset used
        self.new_plot()
        self.tslr_ax = tslr_ax = self.figure.add_subplot(111)
        tslr_ax.set_xlabel(spc.units["x"])
        tslr_ax.set_ylabel(spc.units["y"])
        tslr_ax.hline = tslr_ax.axhline(color="lightgray", lw=width)
        if stack:
            col = cm.get_cmap(colour)
            no = len(spc.y)
            x = spc.x
            for num, y_ in enumerate(spc.y):
                tslr_ax.plot(x, y_, lw=width, color=col(num / no))
        else:
            tslr_ax.plot(spc.x, spc.y, lw=width, color="k")
            axes = [tslr_ax]
            if bars is not None:
                self.bars_ax = bars_ax = tslr_ax.twinx()
                freqs = (
                    bars.wavelengths[0]
                    if spc.genre in ("uv", "ecd")
                    else bars.frequencies[0]
                )
                freqs = freqs + spc.offset
                # show only activities within range requested in calculations
                blade = (freqs >= min(spc.x)) & (freqs <= max(spc.x))
                # but only if there are any in this range
                if blade.any():
                    markerline, stemlines, baseline = bars_ax.stem(
                        freqs[blade],
                        bars.values[0][blade],
                        linefmt="b-",
                        markerfmt=" ",
                        basefmt=" ",
                    )
                    stemlines.set_linewidth(width)
                    bars_ax.set_ylabel(bars.units)
                    bars_ax.tick_params(axis="y", colors="b")
                    axes.append(bars_ax)
            if experimental is not None:
                extremes = [
                    max(abs(max(experimental.y)), abs(min(experimental.y))),
                    max(abs(max(spc.y)), abs(min(spc.y))),
                ]
                if min(extremes) / max(extremes) > 0.1 or not allow_double_axis:
                    # if both will fit fine in one plot
                    tslr_ax.plot(experimental.x, experimental.y, lw=width, color="r")
                else:
                    self.exp_ax = exp_ax = tslr_ax.twinx()
                    exp_ax.plot(experimental.x, experimental.y, lw=width, color="r")
                    exp_ax.spines["left"].set_position(("axes", -0.1))
                    exp_ax.spines["left"].set_visible(True)
                    exp_ax.yaxis.set_ticks_position("left")
                    exp_ax.tick_params(axis="y", colors="r")
                    tslr_ax.yaxis.set_label_coords(-0.17, 0.5)
                    # tslr_ax.tick_params(axis='y', colors='navy')
                    axes.append(exp_ax)
            align_yaxis(axes)
        if reverse_ax:
            tslr_ax.invert_xaxis()
        self.figure.tight_layout()
        self.canvas.draw()

    def change_colour(self, colour):
        if not self.tslr_ax:
            return
        col = cm.get_cmap(colour)
        self.tslr_ax.hline.remove()
        lines = self.tslr_ax.get_lines()
        no = len(lines)
        for num, line in enumerate(lines):
            line.set_color(col(num / no))
        self.tslr_ax.hline = self.tslr_ax.axhline(color="lightgray", lw=0.5)
        self.canvas.draw()



"""Objects representing spectra."""

import logging as lgg
from typing import Dict, Optional, Sequence, Union

import numpy as np

import tesliper  # absolute import to solve problem of circular imports

from .. import datawork as dw
from .array_base import ArrayProperty

# LOGGER
logger = lgg.getLogger(__name__)


class SingleSpectrum:
    """Represents a single spectrum: experimental, averaged from set of conformers, or
    calculated for only one conformer.

    Notes
    -----
    Calling ``len()`` on this class' instance will show a number of data points
    in the spectrum.
    """

    _vibrational_units = {
        "width": "cm-1",
        "start": "cm-1",
        "stop": "cm-1",
        "step": "cm-1",
        "x": "Frequency / cm^(-1)",
    }
    _electronic_units = {
        "width": "eV",
        "start": "nm",
        "stop": "nm",
        "step": "nm",
        "x": "Wavelength / nm",
    }
    _units = {
        "ir": {"y": "Epsilon"},
        "uv": {"y": "Epsilon"},
        "vcd": {"y": "Delta Epsilon"},
        "ecd": {"y": "Delta Epsilon"},
        "raman": {"y": "I(R)+I(L)"},
        "roa": {"y": "I(R)-I(L)"},
    }
    for u in "ir vcd raman roa".split(" "):
        _units[u].update(_vibrational_units)
    for u in ("uv", "ecd"):
        _units[u].update(_electronic_units)
    _spectra_type_ref = dict(
        vcd="vibrational",
        ir="vibrational",
        roa="scattering",
        raman="scattering",
        ecd="electronic",
        uv="electronic",
    )

    def __init__(
        self,
        genre: str,
        values: Sequence[float],
        abscissa: Sequence[float],
        width: float = 0.0,
        fitting: str = "n/a",
        scaling: float = 1.0,
        offset: float = 0.0,
        filenames: Optional[Sequence[str]] = None,
        averaged_by: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        genre : str
            Name of data genre that this object represents.
        values : Sequence[float]
            List of intensity values for each point on the x-axis.
        abscissa : Sequence[float]
            List of x-axis values.
        width : float, optional
            Full width at half maximum used to calculate spectrum, if applies. Provided
            for the record only, by default 0.0.
        fitting : str, optional
            Name of the fitting function used to calculate spectrum, if applies.
            Provided for the record only, by default "n/a".
        scaling : float, optional
            Multiplyier for correction of signal intensity, by default 1.0.
        offset : float, optional
            Correction of the spectrum's shift. Positive value indicates a bathochromic
            shift, negative value indicates a hypsochromic shift. By default 0.0.
        filenames : Optional[Sequence[str]], optional
            List of identifiers of conformers that were used to calculate average
            spectrum, if applies.
        averaged_by : Optional[str], optional
            Energies genre used to calculate average spectrum, if applies.
        """
        self.genre = genre
        self.filenames = [] if filenames is None else filenames
        self.averaged_by = averaged_by
        self.abscissa = abscissa
        self.values = values
        self.start = abscissa[0]
        self.stop = abscissa[-1]
        self.step = abs(abscissa[0] - abscissa[1])
        self.width = width
        self.fitting = fitting
        self.scaling = scaling
        self.offset = offset

    filenames = ArrayProperty(check_against=None, dtype=str)
    abscissa = ArrayProperty(check_against=None)
    values = ArrayProperty(check_against="abscissa")

    @property
    def spectra_type(self):
        """Returns type of spectra: 'vibrational', 'electronic', or 'scattering'."""
        return self._spectra_type_ref[self.genre]

    @property
    def units(self) -> Dict[str, str]:
        """Units in which spectral data is stored. It provides a unit for
        :attr:`.width`, :attr:`.start`, :attr:`.stop`, :attr:`.step`, :attr:`.x`, and
        :attr:`.y`. :attr:`.abscissa` and :attr:`~.SingleSpectrum.values` are stored in
        the same units as :attr:`.x` and :attr:`.y` respectively.
        """
        return self._units[self.genre]

    @property
    def scaling(self) -> Union[int, float]:
        """A factor for correcting the scale of spectra. Setting it to new value changes
        the :attr:`.y` attribute as well. It should be an ``int`` or ``float``.
        """
        return vars(self)["scaling"]

    @scaling.setter
    def scaling(self, factor: Union[int, float]):
        vars(self)["scaling"] = factor
        vars(self)["y"] = self.values * factor

    @property
    def offset(self) -> Union[int, float]:
        """A factor for correcting the shift of spectra. Positive value indicates a
        bathochromic shift, negative value indicates a hypsochromic shift. Setting
        it to new value changes the :attr:`.x` attribute as well. It should be an
        ``int`` or ``float``.
        """
        return vars(self)["offset"]

    @offset.setter
    def offset(self, offset: Union[int, float]):
        vars(self)["offset"] = offset
        vars(self)["x"] = self.abscissa + offset

    @property
    def x(self) -> np.ndarray:
        """Spectra's x-values corrected by adding its :attr:`.offset` to
        :attr:`.abscissa`."""
        return vars(self)["x"]

    @property
    def y(self) -> np.ndarray:
        """Spectra's y-values corrected by multiplying its
        :attr:`~.SingleSpectrum.values` by :attr:`.scaling`."""
        return vars(self)["y"]

    def scale_to(self, spectrum: "SingleSpectrum") -> None:
        """Establishes a scaling factor to best match a scale of the *spectrum* values.

        Parameters
        ----------
        spectrum : SingleSpectrum
            This spectrum's y-axis values will be treated as a reference. If *spectrum*
            has its own scaling factor, it will be taken into account.
        """
        self.scaling = dw.find_scaling(spectrum.y, self.values)

    def shift_to(self, spectrum: "SingleSpectrum") -> None:
        """Establishes an offset factor to best match given *spectrum*.

        Parameters
        ----------
        spectrum : SingleSpectrum
            This spectrum will be treated as a reference. If *spectrum*
            has its own offset factor, it will be taken into account.
        """
        self.offset = dw.find_offset(spectrum.x, spectrum.y, self.abscissa, self.values)

    def __len__(self):
        return len(self.abscissa)


class Spectra(SingleSpectrum):
    """Represents a collection of spectra calculated for a number of conformers.

    .. versionchanged:: 0.9.1
        Corrected ``len()`` behavior.

    Notes
    -----
    Calling ``len()`` on this class' instance will show how many conformers'
    spectra it contains.
    """

    filenames = ArrayProperty(check_against=None, dtype=str)
    abscissa = ArrayProperty(check_against=None)
    values = ArrayProperty(check_against="filenames")

    def __init__(
        self,
        genre: str,
        filenames: Sequence[str],
        values: Sequence[Sequence[float]],
        abscissa: Sequence[float],
        width: float = 0.0,
        fitting: str = "n/a",
        scaling: float = 1.0,
        offset: float = 0.0,
        allow_data_inconsistency: bool = False,
    ):
        """
        Parameters
        ----------
        genre : str
            Name of data genre that this object represents.
        filenames : Optional[Sequence[str]], optional
            List of conformers' identifiers that were used to calculate spectra.
        values : Sequence[float]
            List of intensity values for each point on the x-axis.
        abscissa : Sequence[float]
            List of x-axis values.
        width : float, optional
            Full width at half maximum used to calculate spectra. Provided
            for the record only, by default 0.0.
        fitting : str, optional
            Name of the fitting function used to calculate spectra.
            Provided for the record only, by default "n/a".
        scaling : float, optional
            Multiplyier for correction of signal intensity, by default 1.0.
        offset : float, optional
            Correction of the spectra's shift. Positive value indicates a bathochromic
            shift, negative value indicates a hypsochromic shift. By default 0.0.
        allow_data_inconsistency : bool, optional
            Flag signalizing if instance should allow data inconsistency (see
            :class:`.ArrayPropety` for details).
        """

        self.allow_data_inconsistency = allow_data_inconsistency
        SingleSpectrum.__init__(
            self, genre, values, abscissa, width, fitting, scaling, offset, filenames
        )

    def average(self, energies: "tesliper.glassware.Energies") -> SingleSpectrum:
        """A method for averaging spectra by population of conformers. If this object
        is empty, averaged spectrum will be a flat line at 0.0 intensity.

        Parameters
        ----------
        energies : Energies
            Object with ``populations`` and ``genre`` attributes containing
            respectively: list of populations values as ``numpy.ndarray`` and
            string specifying energy genre.

        Returns
        -------
        SingleSpectrum
            Averaged spectrum.
        """
        populations = energies.populations
        energy_type = energies.genre
        av_spec = dw.calculate_average(self.values, populations)
        if not av_spec.size:
            av_spec = np.zeros(self.abscissa.shape)
        av_spec = SingleSpectrum(
            self.genre,
            av_spec,
            self.abscissa,
            self.width,
            self.fitting,
            scaling=self.scaling,
            offset=self.offset,
            filenames=self.filenames,
            averaged_by=energy_type,
        )
        logger.debug(f"{self.genre} spectrum averaged by {energy_type}.")
        return av_spec

    def scale_to(
        self,
        spectrum: SingleSpectrum,
        average_by: Optional["tesliper.glassware.Energies"] = None,
    ) -> None:
        """Establishes a scaling factor to best match a scale of the *spectrum* values.
        An average spectrum is calculated prior to calculating the factor.
        If *average_by* is given, it is used to average by population of each conformer.
        Otherwise an arithmetic average of spectra is calculated, which may lead
        to inaccurate results.

        Parameters
        ----------
        spectrum : SingleSpectrum
            This spectrum's y-axis values will be treated as a reference. If *spectrum*
            has its own scaling factor, it will be taken into account.
        average_by : Energies, optional
            Energies object, used to calculate average spectrum prior to calculating
            the factor. If not given, a simple arithmetic average of the spectra will
            be calculated.
        """
        if average_by is not None:
            averaged = self.average(energies=average_by)
            averaged.scale_to(spectrum)
            factor = averaged.scaling
        else:
            logger.warning(
                "Trying to find optimal scaling factor for spectra, but no Energies "
                "object given for averaging by population. Results may be inaccurate."
            )
            averaged = np.average(self.values, axis=0)
            factor = dw.find_scaling(spectrum.y, averaged)
        self.scaling = factor

    def shift_to(
        self,
        spectrum: SingleSpectrum,
        average_by: Optional["tesliper.glassware.Energies"] = None,
    ) -> None:
        """Establishes an offset factor to best match given *spectrum*.
        An average spectrum is calculated prior to calculating the factor.
        If *average_by* is given, it is used to average by population of each conformer.
        Otherwise an arithmetic average of spectra is calculated, which may lead
        to inaccurate results.

        Parameters
        ----------
        spectrum : SingleSpectrum
            This spectrum will be treated as a reference. If *spectrum*
            has its own offset factor, it will be taken into account.
        average_by : Energies, optional
            Energies object, used to calculate average spectrum prior to calculating
            the factor. If not given, a simple arithmetic average of the spectra will
            be calculated.
        """
        if average_by is not None:
            averaged = self.average(energies=average_by)
            averaged.shift_to(spectrum)
            factor = averaged.offset
        else:
            logger.warning(
                "Trying to find optimal offset factor for spectra, but no Energies "
                "object given for averaging by population. Results may be inaccurate."
            )
            averaged = np.average(self.values, axis=0)
            factor = dw.find_offset(spectrum.x, spectrum.y, self.abscissa, averaged)
        self.offset = factor

    def __len__(self):
        # must override SingleSpecrum's implementation, because it may have an abscisa
        # but contain no data for conformers
        # e.g. when created in calculations of spectra from an empty activities array
        return len(self.filenames)
